/* Generate ANTLR stuff for VerCors parser*/
lazy val antlrTask = taskKey[Seq[File]]("Generate visitors and listeners from ANTLR grammars")

libraryDependencies += "org.antlr" % "antlr4" % "4.9.2"
libraryDependencies += "org.antlr" % "antlr4-runtime" % "4.9.2"

antlrTask := {
  val cp = (Compile / dependencyClasspath).value.files
  val src = (Compile / sourceDirectory).value / "antlr4"
  val lib = (Compile / unmanagedBase).value
  val target = (Compile / sourceManaged).value / "antlr4" / "generated" / "vercors"
  val log = streams.value.log

  val compileSets: Seq[(java.io.File, Boolean, Set[java.io.File])] = Seq(
    (lib / "LangJavaLexer.g4", false,
      Set(lib / "SpecLexer.g4")),
    (src / "JavaParser.g4", true,
      Set(lib / "SpecParser.g4", lib / "SpecLexer.g4",
        lib / "LangJavaParser.g4", lib / "LangJavaLexer.g4")),
  )

  val allInputFiles: Set[java.io.File] =
    compileSets.foldLeft(Set[java.io.File]()) {
      case (set, (target, _, deps)) => set + target ++ deps
    }

  val cachedCompile = FileFunction.cached(streams.value.cacheDirectory / "antlr4", FilesInfo.hash, FilesInfo.hash) {
    changedSet: Set[File] => {
      for((genTarget, isParser, deps) <- compileSets) {
        val extraArgs = if (isParser) {
          Seq("-listener", "-visitor")
        } else {
          Seq()
        }

        if (changedSet.contains(genTarget) || changedSet.intersect(deps).nonEmpty) {
          val exitCode = scala.sys.process.Process("java", Seq(
            "-cp", Path.makeString(cp),
            "org.antlr.v4.Tool",
            "-o", target.toString,
            "-lib", lib.toString,
            "-package", "generated.vercors",
            genTarget.toString
          ) ++ extraArgs) ! log

          if(exitCode != 0) {
            sys.error(s"Antlr4 failed with exit code $exitCode")
          }
        }
      }

      // Grab all the generated files as our output
      (target ** "*.java").get.toSet ++ (target ** "*.scala").get.toSet
    }
  }

  cachedCompile(allInputFiles).toSeq
}

Compile / sourceGenerators += (Compile / antlrTask).taskValue
Compile / managedSourceDirectories += (Compile / sourceManaged).value / "antlr4"

// Disable documentation generation
Compile / doc / sources := Seq()
Compile / packageDoc / publishArtifact := false