package intermediateRepresentation.syntaxTree

import generic.SourceLocation

trait Expression extends generic.syntaxTree.Expression
                 with Initializer with LambdaBody with IndexExpression

case class NestedInvocationExpression(override val left: Expression,
                                      override val right: Expression,
                                      var srcLocation: SourceLocation,
                                      var precedingWhitespace: String="")
      extends generic.syntaxTree.NestedInvocationExpression with Expression

case class BinaryExpression(override val left: Expression,
                            override val operator: StringNode,
                            override val right: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.BinaryExpression with Expression

case class InstanceofExpression(override val left: Expression,
                                override val right: TypeIndication,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
  extends generic.syntaxTree.InstanceOfExpression with Expression

case class ArrayIndexExpression(override val arrayExpression: Expression,
                                override val indexExpression: IndexExpression,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.ArrayIndexExpression with Expression

trait IndexExpression extends generic.syntaxTree.IndexExpression with IrTreeNode
case class IndexAllExpression(var srcLocation: SourceLocation,
                              var precedingWhitespace: String="")
      extends generic.syntaxTree.IndexAllExpression with Expression

case class RangeExpression(override val start: Expression,
                           override val end: Expression,
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String="")
      extends generic.syntaxTree.RangeExpression with IndexExpression with Expression

case class MethodCallExpression(override val name: StringNode,
                                override val arguments: Arguments,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.MethodCallExpression with Expression

trait NewExpression extends generic.syntaxTree.NewExpression with Expression {
  protected var _typeIndication: TypeIndication
  def typeIndication: TypeIndication = _typeIndication
  def typeIndication_=(x1: generic.syntaxTree.TypeIndication): Unit = {
    if(x1.isIRNode) {
      _typeIndication = x1.asInstanceOf[TypeIndication]
    } else {
      throw new ClassCastException("not an IR node")
    }
  }
}
case class NewObjectExpression (private var _typeArguments: Option[TypeArgumentClause],
                                var _typeIndication: TypeIndication,
                                override val arguments: Arguments,
                                override val body: Option[Body],
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.NewObjectExpression with NewExpression {
  def typeArguments: Option[TypeArgumentClause] = _typeArguments
  def typeArguments_=(x1: Option[generic.syntaxTree.TypeArgumentClause]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isIRNode)) {
      _typeArguments = x1.asInstanceOf[Option[TypeArgumentClause]]
    } else {
      throw new ClassCastException("not an IR node")
    }
  }
}
case class NewArrayExpression(var _typeIndication: TypeIndication,
                              override val dimensions: List[Dimension],
                              override val initializer: Option[ArrayInitializer],
                              var srcLocation: SourceLocation,
                              var precedingWhitespace: String="")
      extends generic.syntaxTree.NewArrayExpression with NewExpression
case class Dimension(override val sizeExpression: Option[Expression],
                     var srcLocation: SourceLocation,
                     var precedingWhitespace: String="")
      extends generic.syntaxTree.Dimension with IrTreeNode

case class CastExpression(override val annotations: List[StringAnnotation],
                          override val typeIndications: List[TypeIndication],
                          override val expression: Expression,
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String="")
      extends generic.syntaxTree.CastExpression with Expression

case class PostfixExpression(override val expression: Expression,
                            override val postfix: StringNode,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.PostfixExpression with Expression

case class PrefixExpression(override val expression: Expression,
                            override val prefix: StringNode,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.PrefixExpression with Expression

case class TernaryExpression (override val condition: Expression,
                              override val thenExpression: Expression,
                              override val elseExpression: Expression,
                              var srcLocation: SourceLocation,
                              var precedingWhitespace: String="")
      extends generic.syntaxTree.TernaryExpression with Expression

case class LambdaExpression(override val hasParentheses: Boolean,
                            override val parameters: List[Parameter],
                            override val body: LambdaBody,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.LambdaExpression with Expression
trait LambdaBody extends generic.syntaxTree.LambdaBody with IrTreeNode

case class MethodReferenceExpression(override val left: Expression,
                                     override val typeArgument: Option[TypeArgumentClause],
                                     override val right: Expression,
                                     var srcLocation: SourceLocation,
                                     var precedingWhitespace: String="")
      extends generic.syntaxTree.MethodReferenceExpression with Expression

case class ParExpression(override val expression: Expression,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="")
      extends generic.syntaxTree.ParExpression with Expression

case class Literal(override val value: StringNode,
                   var srcLocation: SourceLocation,
                   var precedingWhitespace: String = "")
      extends generic.syntaxTree.Literal with Expression

case class ClassName(override val typeIndication: TypeIndication,
                     var srcLocation: SourceLocation,
                     var precedingWhitespace: String="")
      extends generic.syntaxTree.ClassName with Expression

case class GenericInvocationExpression(override val typeArgumentClause: TypeArgumentClause,
                                       override val name: StringNode,
                                       override val arguments: Option[Arguments],
                                       var srcLocation: SourceLocation,
                                       var precedingWhitespace: String="")
      extends generic.syntaxTree.GenericInvocationExpression with Expression
