package languages.Krakatoa

import generated.{JavaLexer, KrakatoaParser}
import intermediateRepresentation.IRProgram
import krakatoa.{DescriptiveErrorListener, TestListenerImpl}
import languages.{IProgram, IProgramGenerator}
import org.antlr.v4.runtime.tree.ParseTreeWalker
import org.antlr.v4.runtime.{CharStream, CharStreams, CommonTokenStream}
import languages.Krakatoa.KrakatoaVisitor
import languages.Krakatoa.syntaxTree.CompilationUnit
import languages.Krakatoa.{FromIRVisitor, ToIRVisitor}

import collection.JavaConverters.asScalaBufferConverter

class KrakatoaProgram(private var rootNode: CompilationUnit) extends IProgram {
  override def toIRProgram: IRProgram = {
    val irAST = new ToIRVisitor().visitCompilationUnit(rootNode)
    //println("\n")
    //println(irAST.toString)
    new IRProgram(irAST)
  }

  override def toText: String = {
    if (rootNode != null) {
      rootNode.toString
    } else {
      "<no code specified>"
    }
  }
}

object KrakatoaProgram extends IProgramGenerator[KrakatoaProgram] {
  override def fromIRProgram(irProgram: IRProgram): KrakatoaProgram = {
    val krakatoaAST = new FromIRVisitor().visitCompilationUnit(irProgram.rootNode)
    KrakatoaProgram(krakatoaAST)
  }

  override def fromFile(filename: String): KrakatoaProgram = {
    val stream = CharStreams.fromFileName(filename)
    fromStream(stream, filename)
  }

  def fromStream(stream: CharStream, filename: String = ""): KrakatoaProgram = {
    val lexer = new JavaLexer(stream)
    val tokens = new CommonTokenStream(lexer)
    val parser = new KrakatoaParser(tokens)
    val parserVisitor = new KrakatoaVisitor(tokens, filename)
    val krakatoaAST = parserVisitor.visitCompilationUnit(parser.compilationUnit())
    KrakatoaProgram(krakatoaAST)
  }
  
  override def fromText(contents: String): KrakatoaProgram = {
    val charstream = CharStreams.fromString(contents);
    val lexer = new JavaLexer(charstream)
    //System.err.println(lexer.getAllTokens.asScala.mkString("\n")) // Prints tokens
    //System.err.flush()
    val tokens = new CommonTokenStream(lexer)
    val parser = new KrakatoaParser(tokens)
    lexer.removeErrorListeners()
    lexer.addErrorListener(DescriptiveErrorListener.INSTANCE)
    parser.removeErrorListeners()
    parser.addErrorListener(DescriptiveErrorListener.INSTANCE)

    //val listener = new TestListenerImpl()
    //parser.compilationUnit().enterRule(listener)

    //val tree = parser.compilationUnit()
    //ParseTreeWalker.DEFAULT.walk(listener, tree)
    //return null

    val visitor = new KrakatoaVisitor(tokens, "")
    val krakatoaAST = visitor.visitCompilationUnit(parser.compilationUnit())
    KrakatoaProgram(krakatoaAST)
  }

}