package languages.Krakatoa.syntaxTree
import generic.syntaxTree.ConstantStringNode
import generic.{SourceLocation, syntaxTree}

trait SpecTypeIndication extends generic.syntaxTree.SpecTypeIndication with TypeIndication
class IntegerSpecTypeIndication(fullTypeString: String = "integer",
                                annotations: List[Annotation] = List.empty,
                                override val typeAsString: String = "integer",
                                srcLocation: SourceLocation,
                                precedingWhitespace: String="")
  extends StringTypeIndication(fullTypeString, annotations, typeAsString, srcLocation, precedingWhitespace) with SpecTypeIndication

object IntegerSpecTypeIndication {
  def unapply(integerSpecTypeIndication: IntegerSpecTypeIndication): Some[(String, List[Annotation], String, SourceLocation, String)] = {
    Some(integerSpecTypeIndication.text,
      integerSpecTypeIndication.annotations,
      integerSpecTypeIndication.typeAsString,
      integerSpecTypeIndication.srcLocation,
      integerSpecTypeIndication.precedingWhitespace)
  }
}
class RealSpecTypeIndication(fullTypeString: String = "real",
                             annotations: List[Annotation] = List.empty,
                             override val typeAsString: String = "real",
                             srcLocation: SourceLocation,
                             precedingWhitespace: String="")
  extends StringTypeIndication(fullTypeString, annotations, typeAsString, srcLocation, precedingWhitespace) with SpecTypeIndication

object RealSpecTypeIndication {
  def unapply(realSpecTypeIndication: RealSpecTypeIndication): Some[(String, List[Annotation], String, SourceLocation, String)] = {
    Some(realSpecTypeIndication.text,
      realSpecTypeIndication.annotations,
      realSpecTypeIndication.typeAsString,
      realSpecTypeIndication.srcLocation,
      realSpecTypeIndication.precedingWhitespace)
  }
}


case class Contract(private var _clauses: List[ContractClause],
                    var srcLocation: SourceLocation,
                    var precedingWhitespace: String = ""
                   ) extends generic.syntaxTree.Contract with KrakatoaTreeNode {
  private var _startSpec: Option[StringNode] = None
  private var _endSpec: Option[StringNode] = None

  def startSpec: Option[StringNode] = _startSpec

  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)){
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }

  def clauses: List[ContractClause] = _clauses

  def clauses_=(x1: List[generic.syntaxTree.ContractClause]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _clauses = x1.asInstanceOf[List[ContractClause]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }

  def endSpec: Option[StringNode] = _endSpec

  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)){
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }

}

class SpecCase(pStartSpec: Option[StringNode],
               var hasAlso: Boolean,
               private var _clauses: List[ContractClause],
               pEndSpec: Option[StringNode],
               var srcLocation: SourceLocation,
               var precedingWhitespace: String = "") extends generic.syntaxTree.SpecCase with ContractClause {
  startSpec = pStartSpec
  endSpec = pEndSpec
  def clauses: List[ContractClause] = _clauses
  def clauses_=(x1: List[generic.syntaxTree.ContractClause]): Unit = {
    if(x1.forall(x => x.isKrakatoaNode)) {
      _clauses = x1.asInstanceOf[List[ContractClause]]
    } else {
      throw new ClassCastException("Not a Krakatoa node")
    }
  }
}
object SpecCase {
  def unapply(specCase: SpecCase): Some[(Option[StringNode], Boolean, List[ContractClause], Option[StringNode], SourceLocation, String)] = {
    Some(specCase.startSpec, specCase.hasAlso, specCase.clauses, specCase.endSpec, specCase.srcLocation, specCase.precedingWhitespace)
  }
}

case class LoopSpecification(override val startSpec: Option[StringNode],
                             override val clauses: List[LoopClause],
                             override val endSpec: Option[StringNode],
                             var srcLocation: SourceLocation,
                             var precedingWhitespace: String="") extends generic.syntaxTree.LoopSpecification with LoopSpecOrLabel

trait LoopSpecOrLabel extends generic.syntaxTree.LoopSpecOrLabel with KrakatoaTreeNode
trait LoopClause extends generic.syntaxTree.LoopClause with KrakatoaTreeNode

case class NonTranslatedLoopClause(var text: String,
                                   var srcLocation: SourceLocation,
                                   var precedingWhitespace: String="")
  extends LoopClause with generic.syntaxTree.NonTranslatedLoopClause

case class NonTranslatedContractClause(var text: String,
                                       var srcLocation: SourceLocation,
                                       var precedingWhitespace: String="")
  extends ContractClause with generic.syntaxTree.NonTranslatedContractClause

trait ContractClause extends generic.syntaxTree.ContractClause with KrakatoaTreeNode {
  private var _startSpec: Option[StringNode] = None
  private var _endSpec: Option[StringNode] = None

  def startSpec: Option[StringNode] = _startSpec

  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)){
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }

  def endSpec: Option[StringNode] = _endSpec

  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)){
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }

}

case class PureModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String = ""
                       ) extends generic.syntaxTree.PureModifier with Modifier
case class GhostModifier(var srcLocation: SourceLocation,
                         var precedingWhitespace: String = ""
                        ) extends generic.syntaxTree.GhostModifier with Modifier
case class SpecMethodModifier(var srcLocation: SourceLocation,
                              var precedingWhitespace: String = ""
                             ) extends generic.syntaxTree.SpecMethodModifier with Modifier {
  override def stringRepresentation: String = "ghost"
}
// AssignsClause is also used for "modifies"-clause
case class AssignsClause(override val assignableKeyword: StringNode = new StringNode("assigns", null),
                            override val expressions: List[Expression],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.AssignableClause with ContractClause with LoopClause

case class AccessibleClause(override val expressions: List[Expression],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.AccessibleClause with ContractClause

case class AssumesClause(val expression: Expression,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String = ""
                        ) extends ContractClause {
  val assumesKeyword: StringNode = new StringNode("assumes", srcLocation)
  val semicolon: StringNode = new StringNode(";", srcLocation)

  override def stringRepresentation: String = {
    assumesKeyword.toString + expression.toString + semicolon.toString
  }
}

case class RequiresClause(override val expression: Expression,
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String = ""
                         ) extends generic.syntaxTree.Precondition with ContractClause

case class EnsuresClause(override val expression: Expression,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String = ""
                        ) extends generic.syntaxTree.Postcondition with ContractClause

case class LoopInvariantClause(override val expression: Expression,
                               var srcLocation: SourceLocation,
                               var precedingWhitespace: String = ""
                              ) extends generic.syntaxTree.LoopInvariant with LoopClause

case class DecreasesClause(override val expression: Expression,
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.MeasuredByClause with ContractClause {
  override val measuredByKeyword = ConstantStringNode("decreases")
  override val predicate = None
}

case class LoopVariantClause(override val expression: Expression,
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String = ""
                          ) extends generic.syntaxTree.DecreasesClause with LoopClause {
  override val decreasesKeyword: syntaxTree.StringNode = StringNode("loop_variant", srcLocation)
}


//TODO: Not sure whether this should support a single contract clause or a list of contract clauses
case class ForClause(val label: StringNode,
                     var contractClause: LoopClause,
                      var srcLocation: SourceLocation,
                     var precedingWhitespace: String = ""
                    ) extends ContractClause with LoopClause {
  val forKeyword: StringNode = new StringNode("for", srcLocation)
  val colon: StringNode = new StringNode(":", srcLocation)

  override def stringRepresentation: String = {
    forKeyword.toString + label.toString + colon.toString + contractClause.toString
  }
}

case class SignalsClause(override val exceptionType: TypeIndication,
                         override val exceptionName: Option[StringNode],
                         override val expression: Some[Expression],
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String = ""
                        ) extends generic.syntaxTree.ExceptionPostcondition with ContractClause


class Behavior(_startSpec: Option[StringNode],
               val behaviorName: StringNode = null,
               _hasAlso: Boolean,
               private var _privacy: Option[StringNode] = None,
               _clauses: List[ContractClause],
               _endSpec: Option[StringNode],
               _srcLocation: SourceLocation,
               _precedingWhitespace: String = ""
              ) extends SpecCase(_startSpec, _hasAlso, _clauses, _endSpec, _srcLocation, _precedingWhitespace) with generic.syntaxTree.Behavior {
  val colon: StringNode = StringNode(":", srcLocation)

  def privacy: Option[StringNode] = _privacy
  def privacy_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)){
      _privacy = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }

  override def stringRepresentation: String = {
    var res = startSpec.getOrElse("").toString
    if (hasAlso) res += alsoKeyword.toString
    res += privacy.getOrElse("").toString
    res += behaviorKeyword.toString
    if (behaviorName != null) res += behaviorName.toString
    res += colon.toString
    res += clauses.mkString
    res += endSpec.getOrElse("").toString
    res
  }
}
object Behavior {
  def unapply(behavior: Behavior): Some[(Option[StringNode], Boolean, Option[StringNode], List[ContractClause], Option[StringNode], SourceLocation, String)] = {
    Some(behavior.startSpec,
      behavior.hasAlso,
      behavior.privacy,
      behavior.clauses,
      behavior.endSpec,
      behavior.srcLocation,
      behavior.precedingWhitespace)
  }
}

class NormalBehavior(startSpec: Option[StringNode],
                     override val behaviorName: StringNode = null,
                     hasAlso: Boolean,
                     privacy: Option[StringNode] = None,
                     override val clauses: List[ContractClause],
                     endSpec: Option[StringNode],
                     srcLocation: SourceLocation,
                     precedingWhitespace : String = ""
                    ) extends Behavior(startSpec, behaviorName, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) with generic.syntaxTree.NormalBehavior
object NormalBehavior {
  def unapply(behavior: NormalBehavior): Some[(Option[StringNode], Boolean, Option[StringNode], List[ContractClause], Option[StringNode], SourceLocation, String)] = {
    Some(behavior.startSpec, behavior.hasAlso, behavior.privacy, behavior.clauses, behavior.endSpec, behavior.srcLocation, behavior.precedingWhitespace)
  }
}

class ExceptionalBehavior(startSpec: Option[StringNode],
                          override val behaviorName: StringNode = null,
                          hasAlso: Boolean,
                          privacy: Option[StringNode] = None,
                          override val clauses: List[ContractClause],
                          endSpec: Option[StringNode],
                          srcLocation: SourceLocation,
                          precedingWhitespace: String = ""
                         ) extends Behavior(startSpec, behaviorName, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace) with generic.syntaxTree.ExceptionalBehavior

object ExceptionalBehavior {
  def unapply(behavior: ExceptionalBehavior): Some[(Option[StringNode], Boolean, Option[StringNode], List[ContractClause], Option[StringNode], SourceLocation, String)] = {
    Some(behavior.startSpec, behavior.hasAlso, behavior.privacy, behavior.clauses, behavior.endSpec, behavior.srcLocation, behavior.precedingWhitespace)
  }
}

case class NonTranslatedStatement(var text: String,
                                  var srcLocation: SourceLocation,
                                  var precedingWhitespace: String="")
  extends Statement with generic.syntaxTree.NonTranslatedStatement

case class AssertStatement(override val content: Expression,
                           override val message: Option[Expression] = None,
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String = ""
                          ) extends generic.syntaxTree.AssertStatement with Statement

case class ForStatement(val label: StringNode,
                        val statement: Statement,
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String = ""
                       ) extends Statement {
  val forKeyword: StringNode = new StringNode("for", srcLocation)
  val colon: StringNode = new StringNode(":", srcLocation)

  override def stringRepresentation: String = {
    forKeyword.toString + label.toString + colon.toString + statement.toString
  }
}

case class GhostStatement(statement: Statement,
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String = ""
                         ) extends generic.syntaxTree.GhostStatement with Statement {
  private var _ghostKeyword: GhostModifier = GhostModifier(srcLocation)

  def ghostKeyword: GhostModifier = _ghostKeyword
  def ghostKeyword_=(x1: generic.syntaxTree.GhostModifier): Unit = {
    x1 match {
      case g: GhostModifier => _ghostKeyword = g
      case _ => throw new ClassCastException("Not a Krakatoa node")
    }
  }
}

case class KrakatoaStatementBlock(private var _startSpec: Option[StringNode],
                                  override val statements: List[Statement],
                                  private var _endSpec: Option[StringNode],
                                  var srcLocation: SourceLocation,
                                  var precedingWhitespace: String="")
  extends generic.syntaxTree.JmlStatementBlock with Statement {
  def startSpec: Option[StringNode] = _startSpec
  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if (x1.isDefined && !x1.get.isInstanceOf[StringNode]) {
      throw new ClassCastException("not a Krakatoa node")
    } else {
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    }
  }
  def endSpec: Option[StringNode] = _endSpec
  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if (x1.isDefined && !x1.get.isInstanceOf[StringNode]) {
      throw new ClassCastException("not a Krakatoa node")
    }else {
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    }
  }
}

case class ResultExpression(var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
  extends generic.syntaxTree.ResultExpression with Expression

case class OldExpression(override val expression: Expression,
                         override val label: Option[StringNode] = None,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String = ""
                        ) extends generic.syntaxTree.OldExpression with Expression

case class AtExpression(override val expression: Expression,
                        override val label: Option[StringNode],
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String = ""
                       ) extends generic.syntaxTree.OldExpression with Expression {
  override val oldKeyword = generic.syntaxTree.ConstantStringNode("\\at")
}

//NOTE: grammar only allows a single expression in a fresh expression
case class FreshExpression(override val expressions: List[Expression],
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String = ""
                          ) extends generic.syntaxTree.FreshExpression with Expression

case class LabeledExpression(name: StringNode,
                             labels: LabelList,
                             expressions: List[Expression],
                             var srcLocation: SourceLocation,
                             var precedingWhitespace: String = ""
                             ) extends Expression {
  val openParenthesis: StringNode = StringNode("(", srcLocation)
  val closeParenthesis: StringNode = StringNode(")", srcLocation)

  override def stringRepresentation: String = {
    var res = name.toString
    res += labels.toString
    res += openParenthesis.toString
    res += expressions.mkString(",")
    res += closeParenthesis.toString
    res
  }
}

trait QuantifiedExpression extends Expression
case class ForallExpression(decls: List[QuantifiedVarDeclaration],
                            expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends QuantifiedExpression {
  val forallKeyword: StringNode = StringNode("\\forall", srcLocation)
  val semi: StringNode = StringNode(";", srcLocation)
  override def stringRepresentation: String = {
    var res = forallKeyword.toString
    res += decls.mkString(",")
    res += semi.toString
    res += expression.toString
    res
  }
}

case class ExistsExpression(decls: List[QuantifiedVarDeclaration],
                            expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace : String = ""
                           ) extends QuantifiedExpression {
  val existsKeyword: StringNode = StringNode("\\exists", srcLocation)
  val semi: StringNode = StringNode(";", srcLocation)
  override def stringRepresentation: String = {
    var res = existsKeyword.toString
    res += decls.mkString(",")
    res += semi.toString
    res += expression.toString
    res
  }
}

case class QuantifiedVarDeclaration(override val declType: TypeIndication,
                               override val declarators: List[StringNode],
                               var srcLocation: SourceLocation,
                               var precedingWhitespace: String = ""
                              ) extends generic.syntaxTree.QuantifiedVarDeclaration with KrakatoaTreeNode {
  override def stringRepresentation: String = {
    var res = declType.toString
    res += declarators.mkString("")
    res
  }
}

//TODO: Skipped [expr]expr for now, not sure if we actually need it in the grammar

case class SpecDeclarationBlock(override val startSpec: Option[StringNode],
                                override val declarations: List[SpecDeclaration],
                                override val endSpec: Option[StringNode],
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="") extends generic.syntaxTree.SpecDeclarationBlock with MemberDeclaration() with TypeDeclaration

case class ModifierBlock(private var _startSpec: Option[StringNode],
                         override val modifiers: List[Modifier],
                         private var _endSpec: Option[StringNode],
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="") extends generic.syntaxTree.JmlModifierBlock with KrakatoaTreeNode with Modifier {
  def startSpec: Option[StringNode] = _startSpec
  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)){
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
  def endSpec: Option[StringNode] = _endSpec
  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isKrakatoaNode)){
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a Krakatoa node")
    }
  }
}

trait SpecDeclaration extends generic.syntaxTree.SpecDeclaration with MemberDeclaration

// Note arguments should allow something like "int a[], real b"
// Parameter allows modifiers which we don't use, but it may NOT support the [] after a variable name
case class PredicateDeclaration(name: StringNode,
                                labels: Option[LabelList],
                                params: List[Parameter],
                                expr: Option[Expression],
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String = ""
                               ) extends generic.syntaxTree.PredicateDeclaration with SpecDeclaration with MemberDeclaration() {
  override val tType: TypeIndication = StringTypeIndication("predicate", Nil, "predicate", srcLocation, precedingWhitespace)

  override def stringRepresentation: String = {
    var res = tType.toString
    res += name.toString
    if (labels.nonEmpty) {
      res += labels.get.toString
    }
    res += openParenthesis.toString
    res += params.mkString(",")
    res += closeParenthesis.toString
    if (expr.nonEmpty) {
      res += equalSign.toString
      res += expr.get.toString
    }
    res += semicolon.toString
    res
  }
}

case class LabelList(labels: List[StringNode],
                     var srcLocation: SourceLocation,
                     var precedingWhitespace: String = ""
                    ) extends KrakatoaTreeNode {
  val openCurlyParenthesis: StringNode = StringNode("{", srcLocation)
  val closeCurlyParenthesis: StringNode = StringNode("}", srcLocation)

  override def stringRepresentation: String = {
    var res = openCurlyParenthesis.toString
    res += labels.mkString(",")
    res += closeCurlyParenthesis.toString
    res
  }
}

case class InductiveDatatypeDeclaration(name: StringNode,
                                        labels: Option[LabelList],
                                        parameters: List[Parameter],
                                        caseDefinitions: List[CaseDefinition],
                                        var srcLocation: SourceLocation,
                                        var precedingWhitespace: String = ""
                                       ) extends SpecDeclaration with MemberDeclaration() {
  val inductiveKeyword: StringNode = StringNode("inductive", srcLocation)
  val openParenthesis: StringNode = StringNode("(", srcLocation)
  val closeParenthesis: StringNode = StringNode(")", srcLocation)
  val openCurlyParenthesis: StringNode = StringNode("{", srcLocation)
  val closeCurlyParenthesis: StringNode = StringNode("}", srcLocation)

  override def stringRepresentation: String = {
    var res = inductiveKeyword.toString
    res += name.toString
    if (labels.nonEmpty) {
      res += labels.get.toString
    }
    res += openParenthesis.toString
    res += parameters.mkString(",")
    res += closeParenthesis.toString
    res += openCurlyParenthesis.toString
    res += caseDefinitions.mkString
    res += closeCurlyParenthesis.toString
    res
  }
}

case class CaseDefinition(name: StringNode,
                          labels: Option[LabelList],
                          expression: Expression,
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String = ""
                         ) extends KrakatoaTreeNode {
  val caseKeyword: StringNode = StringNode("case", srcLocation)
  val colon: StringNode = StringNode(":", srcLocation)
  val semicolon: StringNode = StringNode(";", srcLocation)

  override def stringRepresentation: String = {
    var res = caseKeyword.toString
    res += name.toString
    if (labels.nonEmpty) {
      res += labels.get.toString
    }
    res += colon.toString
    res += expression.toString
    res += semicolon.toString
    res
  }
}

case class InvariantDeclaration(name: StringNode,
                                override val expression: Expression,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String = ""
                               ) extends generic.syntaxTree.InvariantDeclaration with SpecDeclaration with MemberDeclaration() {
  val colon: StringNode = StringNode(":", srcLocation)

  override def stringRepresentation: String = {
    var res = invariantKeyword.toString
    res += name.toString
    res += colon.toString
    res += expression.toString
    res += semicolon.toString
    res
  }
}

case class LemmaDeclaration(name: StringNode,
                            labels: Option[LabelList],
                            expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends SpecDeclaration with MemberDeclaration() {
  val lemmaKeyword: StringNode = StringNode("lemma", srcLocation)
  val colon: StringNode = StringNode(":", srcLocation)
  val semicolon: StringNode = StringNode(";", srcLocation)

  override def stringRepresentation: String = {
    var res = lemmaKeyword.toString
    res += name.toString
    if (labels.nonEmpty) {
      res += labels.get.toString
    }
    res += colon.toString
    res += expression.toString
    res += semicolon.toString
    res
  }
}

case class LogicVariableDeclaration(tType: TypeIndication,
                                    name: StringNode,
                                    labels: Option[LabelList],
                                    parameters: List[Parameter],
                                    expression: Option[Expression],
                                    var srcLocation: SourceLocation,
                                    var precedingWhitespace: String = ""
                                   ) extends SpecDeclaration with MemberDeclaration() {
  val logicKeyword: StringNode = StringNode("logic", srcLocation)
  val openParenthesis: StringNode = StringNode("(", srcLocation)
  val closeParenthesis: StringNode = StringNode(")", srcLocation)
  val equals: StringNode = StringNode("=", srcLocation)
  val semicolon: StringNode = StringNode(";", srcLocation)

  override def stringRepresentation: String = {
    var res = logicKeyword.toString
    res += tType.toString
    res += name.toString
    if (labels.nonEmpty) {
      res += labels.get.toString
    }
    res += openParenthesis.toString
    res += parameters.mkString(",")
    res += closeParenthesis.toString
    if (expression.nonEmpty) {
      res += equals.toString
      res += expression.get.toString
    }
    res += semicolon.toString
    res
  }
}


case class AxiomDeclaration(name: StringNode,
                            labels: Option[LabelList],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.AxiomDeclaration with SpecDeclaration with MemberDeclaration() {
  val colon: StringNode = StringNode(":", srcLocation)

  override def stringRepresentation: String = {
    var res = axiomKeyword.toString
    res += name.toString
    if (labels.nonEmpty) {
      res += labels.get.toString
    }
    res += colon.toString
    res += expression.toString
    res += semicolon.toString
    res
  }
}
case class AxiomaticDeclaration(name: StringNode,
                                declarations: List[SpecDeclaration],
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String = ""
                               ) extends SpecDeclaration with MemberDeclaration() {
  val axiomaticKeyword: StringNode = StringNode("axiomatic", srcLocation)
  val openCurlyParenthesis: StringNode = StringNode("{", srcLocation)
  val closeCurlyParenthesis: StringNode = StringNode("}", srcLocation)

  override def stringRepresentation: String = {
    var res = axiomaticKeyword.toString
    res += name.toString
    res += openCurlyParenthesis.toString
    res += declarations.mkString
    res += closeCurlyParenthesis.toString
    res
  }
}

case class GhostDeclaration(decl: MemberDeclaration,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends SpecDeclaration with generic.syntaxTree.GhostDeclaration with MemberDeclaration() {
  val ghost: GhostModifier = GhostModifier(srcLocation, precedingWhitespace)
}