package languages.OpenJML

import generated.openJML.{JavaParser, JavaPlusLexer}
import intermediateRepresentation.IRProgram
import languages.{IProgram, IProgramGenerator}
import languages.OpenJML.syntaxTree.CompilationUnit
import org.antlr.v4.runtime.{CharStream, CharStreams, CommonTokenStream}

class OpenJMLProgram(private var rootNode: CompilationUnit)
      extends IProgram {

  override def toIRProgram: IRProgram = {
    val irAST = new ToIrVisitor().visitCompilationUnit(rootNode)
    new IRProgram(irAST)
  }

  override def toText: String = {
    if (rootNode != null) {
      rootNode.toString
    } else {
      "<no code specified>"
    }
  }
}

object OpenJMLProgram extends IProgramGenerator[OpenJMLProgram] {
  override def fromIRProgram(irProgram: IRProgram): OpenJMLProgram = {
    val openJMLAST = new FromIRVisitor().visitCompilationUnit(irProgram.rootNode)
    OpenJMLProgram(openJMLAST)
  }

  override def fromText(contents: String): OpenJMLProgram = {
    val charStream = CharStreams.fromString(contents)
    fromStream(charStream)
  }

  def fromFile(filename: String): OpenJMLProgram = {
    val stream = CharStreams.fromFileName(filename)
    fromStream(stream, filename)
  }

  def fromStream(stream: CharStream, filename: String="") : OpenJMLProgram = {
    val lexer = new JavaPlusLexer(stream)
//    println(lexer.getAllTokens)
    val tokens = new CommonTokenStream(lexer)
    val parser = new JavaParser(tokens)

    val parserVisitor = new OpenJMLParserVisitor(tokens, filename)
    val openJMLAST = parserVisitor.visitCompilationUnit(parser.compilationUnit())
    OpenJMLProgram(openJMLAST)
  }

}
