package languages.OpenJML

import generated.openJML.JavaParserBaseVisitor
import org.antlr.v4.runtime.{CommonTokenStream, ParserRuleContext}
import org.antlr.v4.runtime.tree.TerminalNode

import scala.jdk.CollectionConverters.*

/**
 * Generate String from parseTree, *including hidden tokens before nodes*
 * @param tokens: CommonTokenStream to get hidden tokens
 */
class ParserPrintVisitor(tokens: CommonTokenStream) extends JavaParserBaseVisitor[String] {
  override def visitTerminal(node: TerminalNode): String = {
    var res = node.toString
    val idx = node.getSymbol.getTokenIndex
    val hiddenTokens = tokens.getHiddenTokensToLeft(idx).asScala
    if (hiddenTokens != null) {
      res = hiddenTokens.map(_.getText).mkString + res
    }
    res
  }

  override def defaultResult(): String = ""

  override def aggregateResult(aggregate: String, nextResult: String): String = {
    aggregate + nextResult
  }

  def visitSkipFirstWS(node: TerminalNode): String = {
    val res = visit(node)
    val hidden = tokens.getHiddenTokensToLeft(node.getSymbol.getTokenIndex).asScala
    if(hidden != null) {
      val hiddenLen = hidden.map(x => x.getText).mkString.length
      res.substring(hiddenLen)
    } else {
      res
    }
  }
  def visitSkipFirstWS(node: ParserRuleContext): String = {
    val res = visit(node)
    val hidden = tokens.getHiddenTokensToLeft(node.start.getTokenIndex).asScala
    if(hidden != null) {
      val hiddenLen = hidden.map(x => x.getText).mkString.length
      res.substring(hiddenLen)
    } else {
      res
    }
  }

}
