package languages.OpenJML.syntaxTree
import generic.SourceLocation
import generic.syntaxTree.NodeType

trait SpecTypeIndication extends generic.syntaxTree.SpecTypeIndication with TypeIndication
class IntegerSpecTypeIndication(fullTypeString: String = "\\bigint",
                                annotations: List[Annotation] = List.empty,
                                override val typeAsString: String = "\\bigint",
                                srcLocation: SourceLocation,
                                precedingWhitespace: String="")
  extends StringTypeIndication(fullTypeString, annotations, typeAsString, srcLocation, precedingWhitespace) with SpecTypeIndication

object IntegerSpecTypeIndication {
  def unapply(integerSpecTypeIndication: IntegerSpecTypeIndication): Some[(String, List[Annotation], String, SourceLocation, String)] = {
    Some(integerSpecTypeIndication.text,
      integerSpecTypeIndication.annotations,
      integerSpecTypeIndication.typeAsString,
      integerSpecTypeIndication.srcLocation,
      integerSpecTypeIndication.precedingWhitespace)
  }
}
class RealSpecTypeIndication(fullTypeString: String = "\\real",
                             annotations: List[Annotation] = List.empty,
                             override val typeAsString: String = "\\real",
                             srcLocation: SourceLocation,
                             precedingWhitespace: String="")
  extends StringTypeIndication(fullTypeString, annotations, typeAsString, srcLocation, precedingWhitespace) with SpecTypeIndication

object RealSpecTypeIndication {
  def unapply(realSpecTypeIndication: RealSpecTypeIndication): Some[(String, List[Annotation], String, SourceLocation, String)] = {
    Some(realSpecTypeIndication.text,
      realSpecTypeIndication.annotations,
      realSpecTypeIndication.typeAsString,
      realSpecTypeIndication.srcLocation,
      realSpecTypeIndication.precedingWhitespace)
  }
}
case class JmlModifierBlock(private var _startSpec: Option[StringNode],
                            override val modifiers: List[Modifier],
                            private var _endSpec: Option[StringNode],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.JmlModifierBlock with Modifier {
  def startSpec: Option[StringNode] = _startSpec
  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isOpenJMLNode)) {
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not an OpenJML node")
    }
  }
  def endSpec: Option[StringNode] = _endSpec
  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isOpenJMLNode)) {
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not an OpenJML node")
    }
  }
}

case class PureModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.PureModifier with Modifier
case class NullableModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.NullableModifier with Modifier
case class NullableByDefaultModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.NullableByDefaultModifier with Modifier
case class NonnullModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.NonnullModifier with Modifier
case class GhostModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.GhostModifier with Modifier
class SpecMethodModifier(srcLocation: SourceLocation,
                              precedingWhitespace: String="")
      extends ModelModifier(srcLocation, precedingWhitespace) 
        with generic.syntaxTree.SpecMethodModifier
object SpecMethodModifier {
  def unapply(specMethodModifier: SpecMethodModifier): Some[(SourceLocation, String)] = {
    Some(specMethodModifier.srcLocation, specMethodModifier.precedingWhitespace)
  }
}
case class HelperModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.HelperModifier with Modifier
case class MonitoredModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.MonitoredModifier with Modifier
class ModelModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.ModelModifier with Modifier
object ModelModifier {
  def unapply(modelModifier: ModelModifier): Some[(SourceLocation, String)] = {
    Some(modelModifier.srcLocation, modelModifier.precedingWhitespace)
  }
}
case class InstanceModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.InstanceModifier with Modifier
class SpecPublicModifier(_srcLocation: SourceLocation,
                              _precedingWhitespace: String="")
      extends ModifierStringNode("spec_public", _srcLocation, _precedingWhitespace)
object SpecPublicModifier {
  def unapply(modifier: SpecPublicModifier): Some[(SourceLocation,
                                                    String)] = {
    Some(modifier.srcLocation,
          modifier.precedingWhitespace)
  }
}
class SpecProtectedModifier(_srcLocation: SourceLocation,
                              _precedingWhitespace: String="")
      extends ModifierStringNode("spec_protected", _srcLocation, _precedingWhitespace)
object SpecProtectedModifier {
  def unapply(modifier: SpecProtectedModifier): Some[(SourceLocation,
                                                    String)] = {
    Some(modifier.srcLocation,
          modifier.precedingWhitespace)
  }
}

case class Contract(private var _clauses: List[ContractClause],
                    var srcLocation: SourceLocation,
                    var precedingWhitespace: String="")
      extends generic.syntaxTree.Contract with OpenJMLTreeNode {
  private var _startSpec: Option[StringNode] = None
  private var _endSpec: Option[StringNode] = None

  def startSpec: Option[StringNode] = _startSpec
  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isOpenJMLNode)) {
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not an OpenJML node")
    }
  }
  def clauses: List[ContractClause] = _clauses
  def clauses_=(x1: List[generic.syntaxTree.ContractClause]): Unit = {
    if(x1.forall(x => x.isOpenJMLNode)) {
      _clauses = x1.asInstanceOf[List[ContractClause]]
    } else {
      throw new ClassCastException("not an OpenJML node")
    }
  }
  def endSpec: Option[StringNode] = _endSpec
  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isOpenJMLNode)) {
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not an OpenJML node")
    }
  }
}

class SpecCase(pStartSpec: Option[StringNode],
               var hasAlso: Boolean,
               private var _clauses: List[ContractClause],
               pEndSpec: Option[StringNode],
               var srcLocation: SourceLocation,
               var precedingWhitespace: String="")
      extends generic.syntaxTree.SpecCase with ContractClause {
  startSpec = pStartSpec
  endSpec = pEndSpec
  def clauses: List[ContractClause] = _clauses
  def clauses_=(x1: List[generic.syntaxTree.ContractClause]): Unit = {
    if(x1.forall(x => x.isOpenJMLNode)) {
      _clauses = x1.asInstanceOf[List[ContractClause]]
    } else {
      throw new ClassCastException("not an OpenJML node")
    }
  }
}
object SpecCase {
  def unapply(specCase: SpecCase): Some[(Option[StringNode],
                                          Boolean,
                                          List[ContractClause],
                                          Option[StringNode],
                                          SourceLocation,
                                          String)] = {
    Some(specCase.startSpec,
          specCase.hasAlso,
          specCase.clauses,
          specCase.endSpec,
          specCase.srcLocation,
          specCase.precedingWhitespace)
  }
}

case class LoopSpecification(override val startSpec: Option[StringNode],
                             override val clauses: List[LoopClause],
                             override val endSpec: Option[StringNode],
                             var srcLocation: SourceLocation,
                             var precedingWhitespace: String="")
      extends generic.syntaxTree.LoopSpecification with LoopSpecOrLabel

trait LoopSpecOrLabel extends generic.syntaxTree.LoopSpecOrLabel with OpenJMLTreeNode

trait ContractClause extends generic.syntaxTree.ContractClause with OpenJMLTreeNode {
  private var _startSpec: Option[StringNode] = None
  private var _endSpec: Option[StringNode] = None
  def startSpec: Option[StringNode] = _startSpec
  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isOpenJMLNode)) {
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not an OpenJML node")
    }
  }
  def endSpec: Option[StringNode] = _endSpec
  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isOpenJMLNode)) {
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not an OpenJML node")
    }
  }
}
trait LoopClause extends generic.syntaxTree.LoopClause with OpenJMLTreeNode

case class NonTranslatedContractClause(var text: String,
                                       var srcLocation: SourceLocation,
                                       var precedingWhitespace: String="")
  extends ContractClause with generic.syntaxTree.NonTranslatedContractClause

case class NonTranslatedLoopClause(var text: String,
                                       var srcLocation: SourceLocation,
                                       var precedingWhitespace: String="")
  extends LoopClause with generic.syntaxTree.NonTranslatedLoopClause

case class AssignableClause(override val expressions: List[Expression],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.AssignableClause with ContractClause

case class AccessibleClause(override val expressions: List[Expression],
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String="")
      extends generic.syntaxTree.AccessibleClause with ContractClause

case class ForallVariableDeclaration(override val declaration: QuantifiedVarDeclaration,
                                     var srcLocation: SourceLocation,
                                     var precedingWhitespace: String="")
      extends generic.syntaxTree.ForallVariableDeclaration with ContractClause

case class OldVariableDeclaration(override val modifiers: List[Modifier],
                                  override val typeIndication: TypeIndication,
                                  override val declarators: List[VariableDeclarator],
                                  var srcLocation: SourceLocation,
                                  var precedingWhitespace: String="")
      extends generic.syntaxTree.OldVariableDeclaration with ContractClause

case class Precondition(override val expression: Expression,
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.Precondition with ContractClause

case class Postcondition(override val expression: Expression,
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String="")
      extends generic.syntaxTree.Postcondition with ContractClause

case class DivergeClause(override val expression: Expression,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="")
      extends generic.syntaxTree.DivergeClause with ContractClause

case class LoopInvariant(override val expression: Expression,
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.LoopInvariant with LoopClause

case class LoopModifiesClause(override val expressions: List[Expression],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.LoopModifiesClause with LoopClause

case class DecreasesClause(override val expression: Expression,
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.DecreasesClause with LoopClause

case class MeasuredByClause(override val expression: Expression,
                            override val predicate: Option[Expression],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.MeasuredByClause with ContractClause

case class ExceptionalPostcondition(override val exceptionType: TypeIndication,
                                    override val exceptionName: Option[StringNode],
                                    override val expression: Option[Expression],
                                    var srcLocation: SourceLocation,
                                    var precedingWhitespace: String="")
      extends generic.syntaxTree.ExceptionPostcondition with ContractClause

case class SignalsOnlyClause(override val exceptionTypes: List[TypeIndication],
                             var srcLocation: SourceLocation,
                             var precedingWhitespace: String="")
      extends generic.syntaxTree.SignalsOnlyClause with ContractClause

class Behavior(_startSpec: Option[StringNode],
               _hasAlso: Boolean,
               private var _privacy: Option[StringNode],
               _clauses: List[ContractClause],
               _endSpec: Option[StringNode],
               _srcLocation: SourceLocation,
               _precedingWhitespace: String="")
      extends SpecCase(_startSpec, _hasAlso, _clauses, _endSpec, _srcLocation, _precedingWhitespace)
          with generic.syntaxTree.Behavior  {
  def privacy: Option[StringNode] = _privacy
  def privacy_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isOpenJMLNode)) {
      _privacy = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not an OpenJML node")
    }
  }
}
object Behavior {
  def unapply(behavior: Behavior): Some[(Option[StringNode],
                                          Boolean,
                                          Option[StringNode],
                                          List[ContractClause],
                                          Option[StringNode],
                                          SourceLocation,
                                          String)] = {
    Some(behavior.startSpec,
          behavior.hasAlso,
          behavior.privacy,
          behavior.clauses,
          behavior.endSpec,
          behavior.srcLocation,
          behavior.precedingWhitespace)
  }
}
class NormalBehavior(startSpec: Option[StringNode],
                     hasAlso: Boolean,
                     privacy: Option[StringNode],
                     clauses: List[ContractClause],
                     endSpec: Option[StringNode],
                     srcLocation: SourceLocation,
                     precedingWhitespace: String="")
      extends Behavior(startSpec,hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace)
        with generic.syntaxTree.NormalBehavior
object NormalBehavior {
  def unapply(behavior: NormalBehavior): Some[(Option[StringNode],
                                                Boolean,
                                                Option[StringNode],
                                                List[ContractClause],
                                                Option[StringNode],
                                                SourceLocation,
                                                String)] = {
    Some(behavior.startSpec,
          behavior.hasAlso,
          behavior.privacy,
          behavior.clauses,
          behavior.endSpec,
          behavior.srcLocation,
          behavior.precedingWhitespace)
  }
}
class ExceptionalBehavior(startSpec: Option[StringNode],
                          hasAlso: Boolean,
                          privacy: Option[StringNode],
                          clauses: List[ContractClause],
                          endSpec: Option[StringNode],
                          srcLocation: SourceLocation,
                          precedingWhitespace: String="")
      extends Behavior(startSpec, hasAlso, privacy, clauses, endSpec, srcLocation, precedingWhitespace)
        with generic.syntaxTree.ExceptionalBehavior
object ExceptionalBehavior {
  def unapply(behavior: ExceptionalBehavior): Some[(Option[StringNode],
                                                    Boolean,
                                                    Option[StringNode],
                                                    List[ContractClause],
                                                    Option[StringNode],
                                                    SourceLocation,
                                                    String)] = {
    Some(behavior.startSpec,
          behavior.hasAlso,
          behavior.privacy,
          behavior.clauses,
          behavior.endSpec,
          behavior.srcLocation,
          behavior.precedingWhitespace)
  }
}

case class NonTranslatedStatement(var text: String,
                                  var srcLocation: SourceLocation,
                                  var precedingWhitespace: String="")
  extends generic.syntaxTree.NonTranslatedStatement with Statement

case class GhostStatement(statement: Statement,
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String = ""
                         ) extends generic.syntaxTree.GhostStatement with Statement {
  private var _ghostKeyword: GhostModifier = GhostModifier(srcLocation)
  def ghostKeyword: GhostModifier = _ghostKeyword

  def ghostKeyword_=(x1: generic.syntaxTree.GhostModifier): Unit = {
    x1 match {
      case g: GhostModifier => _ghostKeyword = g
      case _ => throw new ClassCastException("Not an OpenJML node")
    }
  }
}
case class AssertStatement (override val content: Expression,
                            override val message: Option[Expression],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.AssertStatement with Statement
case class AssumeStatement (override val content: Expression,
                            override val message: Option[Expression],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.AssumeStatement with Statement
case class UnreachableStatement(var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.UnreachableStatement with Statement
case class SetStatement(override val assignment: Expression,
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.SetStatement with Statement

case class JmlStatementBlock(private var _startSpec: Option[StringNode],
                             override val statements: List[Statement],
                             private var _endSpec: Option[StringNode],
                             var srcLocation: SourceLocation,
                             var precedingWhitespace: String="")
      extends generic.syntaxTree.JmlStatementBlock with Statement {
  def startSpec: Option[StringNode] = _startSpec
  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if (x1.isDefined && !x1.get.isInstanceOf[StringNode]) {
      throw new ClassCastException("not an OpenJML node")
    }else {
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    }
//    x1 match {
//      case start: Option[StringNode] => startSpec = start
//      case _ => throw new ClassCastException("not an OpenJML node")
//    }
  }
  def endSpec: Option[StringNode] = _endSpec
  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if (x1.isDefined && !x1.get.isInstanceOf[StringNode]) {
      throw new ClassCastException("not an OpenJML node")
    }else {
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    }
//    x1 match {
////      case end: Option[StringNode] => endSpec = end
////      case _ => throw new ClassCastException("not an OpenJML node")
//      case None => endSpec = None
//      case Some(value) => value match {
//        case str: StringNode => endSpec = Some(str)
//        case _ => throw new ClassCastException("not an OpenJML node")
//      }
//    }
  }
}

case class ResultExpression(var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
  extends generic.syntaxTree.ResultExpression with Expression

case class OldExpression(override val expression: Expression,
                         override val label: Option[StringNode],
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="")
      extends generic.syntaxTree.OldExpression with Expression

case class PreExpression(override val expression: Expression,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="")
      extends generic.syntaxTree.PreExpression with Expression

case class NotAssignedExpression(override val storeRefs: List[Expression],
                                 var srcLocation: SourceLocation,
                                 var precedingWhitespace: String="")
      extends generic.syntaxTree.NotAssignedExpression with Expression

case class NotModifiedExpression(override val storeRefs: List[Expression],
                                 var srcLocation: SourceLocation,
                                 var precedingWhitespace: String="")
      extends generic.syntaxTree.NotModifiedExpression with Expression

case class OnlyAccessedExpression(override val storeRefs: List[Expression],
                                 var srcLocation: SourceLocation,
                                 var precedingWhitespace: String="")
      extends generic.syntaxTree.OnlyAccessedExpression with Expression

case class OnlyAssignedExpression(override val storeRefs: List[Expression],
                                 var srcLocation: SourceLocation,
                                 var precedingWhitespace: String="")
      extends generic.syntaxTree.OnlyAssignedExpression with Expression

case class FreshExpression(override val expressions: List[Expression],
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String="")
      extends generic.syntaxTree.FreshExpression with Expression

case class NonNullElementsExpression(override val expression: Expression,
                                 var srcLocation: SourceLocation,
                                 var precedingWhitespace: String="")
      extends generic.syntaxTree.NonNullElementsExpression with Expression

case class TypeOfExpression(override val expression: Expression,
                                 var srcLocation: SourceLocation,
                                 var precedingWhitespace: String="")
      extends generic.syntaxTree.TypeOfExpression with Expression

case class TypeExpression(override val typeIndication: TypeIndication,
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String="")
      extends generic.syntaxTree.TypeExpression with Expression

case class InformalDescription(text: String,
                               var srcLocation: SourceLocation,
                               var precedingWhitespace: String="")
  extends Expression {
  override def stringRepresentation: String = text
}

trait QuantifiedExpression extends generic.syntaxTree.QuantifiedExpression with Expression {
  val _declaration: QuantifiedVarDeclaration
  override val declaration = List(_declaration)
}
case class ForallExpression(override val _declaration: QuantifiedVarDeclaration,
                            override val guard: Option[Expression],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.ForallExpression with QuantifiedExpression
case class ExistsExpression(override val _declaration: QuantifiedVarDeclaration,
                            override val guard: Option[Expression],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.ExistsExpression with QuantifiedExpression
case class MaxQuantifiedExpression(override val _declaration: QuantifiedVarDeclaration,
                            override val guard: Option[Expression],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.MaxQuantifiedExpression with QuantifiedExpression
case class MinExpression(override val _declaration: QuantifiedVarDeclaration,
                            override val guard: Option[Expression],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.MinExpression with QuantifiedExpression
case class NumOfExpression(override val _declaration: QuantifiedVarDeclaration,
                            override val guard: Option[Expression],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.NumOfExpression with QuantifiedExpression
case class ProductExpression(override val _declaration: QuantifiedVarDeclaration,
                            override val guard: Option[Expression],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.ProductExpression with QuantifiedExpression
case class SumExpression(override val _declaration: QuantifiedVarDeclaration,
                            override val guard: Option[Expression],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.SumExpression with QuantifiedExpression


class QuantifiedVarDeclaration(override val declType: TypeIndication,
                                    override val declarators: List[StringNode],
                                    var srcLocation: SourceLocation,
                                    var precedingWhitespace: String="")
      extends generic.syntaxTree.QuantifiedVarDeclaration with OpenJMLTreeNode
object QuantifiedVarDeclaration {
  def unapply(varDecl: QuantifiedVarDeclaration): Some[(TypeIndication,
                                                        List[StringNode],
                                                        SourceLocation,
                                                        String)] = {
    Some(varDecl.declType,
          varDecl.declarators,
          varDecl.srcLocation,
          varDecl.precedingWhitespace)
  }
}
class NonNullQuantifiedVarDeclaration(override val declType: TypeIndication,
                                    override val declarators: List[StringNode],
                                    srcLocation: SourceLocation,
                                    precedingWhitespace: String="")
      extends QuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace)
        with generic.syntaxTree.NonNullQuantifiedVarDeclaration
object NonNullQuantifiedVarDeclaration {
  def unapply(varDecl: NonNullQuantifiedVarDeclaration): Some[(TypeIndication,
                                                                List[StringNode],
                                                                SourceLocation,
                                                                String)] = {
    Some(varDecl.declType,
          varDecl.declarators,
          varDecl.srcLocation,
          varDecl.precedingWhitespace)
  }
}
class NullableQuantifiedVarDeclaration(override val declType: TypeIndication,
                                    override val declarators: List[StringNode],
                                    srcLocation: SourceLocation,
                                    precedingWhitespace: String="")
      extends QuantifiedVarDeclaration(declType, declarators, srcLocation, precedingWhitespace)
    with generic.syntaxTree.NullableQuantifiedVarDeclaration
object NullableQuantifiedVarDeclaration {
  def unapply(varDecl: NullableQuantifiedVarDeclaration): Some[(TypeIndication,
                                                                List[StringNode],
                                                                SourceLocation,
                                                                String)] = {
    Some(varDecl.declType,
          varDecl.declarators,
          varDecl.srcLocation,
          varDecl.precedingWhitespace)
  }
}

case class SetComprehension(var _typeIndication: TypeIndication,
                            override val declaration: QuantifiedVarDeclaration,
                            override val containsExpression: Expression,
                            override val filterExpression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.SetComprehension with NewExpression

class ImpliesOperator (pSrcLocation: SourceLocation,
                       pPrecedingWhitespace: String="")
  extends StringNode("==>", pSrcLocation, pPrecedingWhitespace)
object ImpliesOperator{
  def unapply(op: ImpliesOperator): Some[(SourceLocation, String)] = {
    Some(op.srcLocation, op.precedingWhitespace)
  }
}

case class SpecDeclarationBlock(override val startSpec: Option[StringNode],
                                override val declarations: List[SpecDeclaration],
                                override val endSpec: Option[StringNode],
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.SpecDeclarationBlock with MemberDeclaration() with TypeDeclaration
trait SpecDeclaration extends generic.syntaxTree.SpecDeclaration with MemberDeclaration
case class InvariantDeclaration(override val expression: Expression,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.InvariantDeclaration with SpecDeclaration with MemberDeclaration()
case class InitiallyDeclaration(override val expression: Expression,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.InitiallyDeclaration with SpecDeclaration with MemberDeclaration()
case class ReadableDeclaration(override val identifier: StringNode,
                                override val expression: Expression,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.ReadableDeclaration with SpecDeclaration with MemberDeclaration()
case class WritableDeclaration(override val identifier: StringNode,
                                override val expression: Expression,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.WritableDeclaration with SpecDeclaration with MemberDeclaration()
case class MonitorsForDeclaration(override val identifier: StringNode,
                                override val expressions: List[Expression],
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.MonitorsForDeclaration with SpecDeclaration with MemberDeclaration()
case class AxiomDeclaration(override val expression: Expression,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.AxiomDeclaration with SpecDeclaration with MemberDeclaration()

case class GhostDeclaration(override val ghost: GhostModifier,
                            override val decl: MemberDeclaration,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = ""
                           ) extends generic.syntaxTree.GhostDeclaration with SpecDeclaration with MemberDeclaration() {
  override def stringRepresentation: String = {
    decl.modifiers = decl.modifiers.prepended(ghost)
    val res = decl.toString
    decl.modifiers = decl.modifiers.tail
    res
  }
}

case class ModelDeclaration(override val model: ModelModifier,
                            override val decl: MemberDeclaration,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = "")
      extends generic.syntaxTree.ModelDeclaration with SpecDeclaration with MemberDeclaration() {
  override def stringRepresentation: String = {
    decl.modifiers = decl.modifiers.prepended(model)
    val res = decl.toString
    decl.modifiers = decl.modifiers.tail
    res
  }
}

case class NonTranslatedSpecDeclaration(var text: String,
                                        var srcLocation: SourceLocation,
                                        var precedingWhitespace: String = ""
                                       ) extends generic.syntaxTree.NonTranslatedSpecDeclaration with SpecDeclaration with MemberDeclaration()

case class SpecMethodDeclaration(pName: StringNode,
                                 pReturnType: Option[TypeIndication],
                                 pParameters: List[Parameter],
                                 pBody: BlockOrSemicolon,
                                 pSrcLocation: SourceLocation,
                                 pPrecedingWhitespace: String="")
  extends MethodDeclaration(pName, pReturnType, pParameters, None,
                            None, None, None, pBody,
                            pSrcLocation, pPrecedingWhitespace)
    with SpecDeclaration with generic.syntaxTree.SpecMethodDeclaration
