package languages.VerCors

import generated.vercors.{JavaParser, LangJavaLexer}
import intermediateRepresentation.IRProgram
import languages.{IProgram, IProgramGenerator}
import languages.VerCors.syntaxTree.CompilationUnit
import org.antlr.v4.runtime.{CharStream, CharStreams, CommonTokenStream}
import parserScala.TestListenerImpl
import parserJava.DescriptiveErrorListener

class VerCorsProgram(private var rootNode: CompilationUnit)
      extends IProgram {
  override def toIRProgram: IRProgram = IRProgram(ToIrVisitor().visitCompilationUnit(rootNode))

  override def toText: String = {
    if (rootNode != null) {
      rootNode.toString
    } else {
      "<no code specified>"
    }
  }
}

object VerCorsProgram extends IProgramGenerator[VerCorsProgram] {
  override def fromIRProgram(irProgram: IRProgram): VerCorsProgram = {
    val root = new FromIRVisitor().visitCompilationUnit(irProgram.rootNode)
    VerCorsProgram(root)
  }

  def fromTextWithListener(contents: String): VerCorsProgram = {
    val charstream = CharStreams.fromString(contents);
    val lexer = new LangJavaLexer(charstream)
    val tokens = new CommonTokenStream(lexer)
    val parser = new JavaParser(tokens)
    lexer.removeErrorListeners()
    lexer.addErrorListener(DescriptiveErrorListener.INSTANCE)
    parser.removeErrorListeners()
    parser.addErrorListener(DescriptiveErrorListener.INSTANCE)

    val listener = new TestListenerImpl()
    parser.compilationUnit().enterRule(listener)
    return null
  }
  override def fromText(contents: String): VerCorsProgram = {
    val charStream = CharStreams.fromString(contents)
    fromStream(charStream)
  }

  def fromFile(filename: String): VerCorsProgram = {
    val stream = CharStreams.fromFileName(filename)
    fromStream(stream, filename)
  }

  def fromStream(stream: CharStream, filename: String="") : VerCorsProgram = {
    val lexer = new LangJavaLexer(stream)
    val tokens = new CommonTokenStream(lexer)
//    println(lexer.getAllTokens)
    val parser = new JavaParser(tokens)

    val parserVisitor = new VerCorsParserVisitor(tokens, filename)
    val vercorsAST = parserVisitor.visitCompilationUnit(parser.compilationUnit())
    VerCorsProgram(vercorsAST)
  }
}
