package languages.VerCors.syntaxTree

import generic.syntaxTree.ConstantStringNode
import generic.{SourceLocation, syntaxTree}

trait SpecTypeIndication extends generic.syntaxTree.SpecTypeIndication with TypeIndication
class IntegerSpecTypeIndication(fullTypeString: String = "int",
                                annotations: List[Annotation] = List.empty,
                                override val typeAsString: String = "int",
                                srcLocation: SourceLocation,
                                precedingWhitespace: String="")
  extends StringTypeIndication(fullTypeString, annotations, typeAsString, srcLocation, precedingWhitespace) with SpecTypeIndication

object IntegerSpecTypeIndication {
  def unapply(integerSpecTypeIndication: IntegerSpecTypeIndication): Some[(String, List[Annotation], String, SourceLocation, String)] = {
    Some(integerSpecTypeIndication.text,
      integerSpecTypeIndication.annotations,
      integerSpecTypeIndication.typeAsString,
      integerSpecTypeIndication.srcLocation,
      integerSpecTypeIndication.precedingWhitespace)
  }
}
case class JmlModifierBlock(private var _startSpec: Option[StringNode],
                            override val modifiers: List[Modifier],
                            private var _endSpec: Option[StringNode],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.JmlModifierBlock with Modifier {
  def startSpec: Option[StringNode] = _startSpec
  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isVerCorsNode)) {
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a VerCors node")
    }
  }
  def endSpec: Option[StringNode] = _endSpec
  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isVerCorsNode)) {
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a VerCors node")
    }
  }
}

case class PureModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.PureModifier with Modifier
class GhostModifier(var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.GhostModifier with Modifier
object GhostModifier {
  def unapply(ghostModifier: GhostModifier): Some[(SourceLocation, String)] = {
    Some(ghostModifier.srcLocation, ghostModifier.precedingWhitespace)
  }
}
class SpecMethodModifier(srcLocation: SourceLocation,
                         precedingWhitespace: String="")
      extends GhostModifier(srcLocation, precedingWhitespace)
        with generic.syntaxTree.SpecMethodModifier
object SpecMethodModifier {
  def unapply(specMethodModifier: SpecMethodModifier): Some[(SourceLocation, String)] = {
    Some(specMethodModifier.srcLocation, specMethodModifier.precedingWhitespace)
  }
}
class InlineModifier(_srcLocation: SourceLocation,
                     _precedingWhitespace: String="")
  extends ModifierStringNode("inline", _srcLocation, _precedingWhitespace)
object InlineModifier {
  def unapply(modifier: InlineModifier): Some[(SourceLocation,
                                                String)] = {
    Some(modifier.srcLocation,
          modifier.precedingWhitespace)
  }
}
class ThreadLocalModifier(_srcLocation: SourceLocation,
                          _precedingWhitespace: String="")
  extends ModifierStringNode("thread_local", _srcLocation, _precedingWhitespace)
object ThreadLocalModifier {
  def unapply(modifier: ThreadLocalModifier): Some[(SourceLocation,
                                                     String)] = {
    Some(modifier.srcLocation,
          modifier.precedingWhitespace)
  }
}

case class Contract(private var _clauses: List[ContractClause],
                    var srcLocation: SourceLocation,
                    var precedingWhitespace: String="")
  extends generic.syntaxTree.Contract with VerCorsTreeNode {
  private var _startSpec: Option[StringNode] = None
  private var _endSpec: Option[StringNode] = None

  def startSpec: Option[StringNode] = _startSpec
  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isVerCorsNode)) {
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a VerCors node")
    }
  }
  def clauses: List[ContractClause] = _clauses
  def clauses_=(x1: List[generic.syntaxTree.ContractClause]): Unit = {
    if(x1.forall(x => x.isVerCorsNode)) {
      _clauses = x1.asInstanceOf[List[ContractClause]]
    } else {
      throw new ClassCastException("not a VerCors node")
    }
  }
  def endSpec: Option[StringNode] = _endSpec
  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isVerCorsNode)) {
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a VerCors node")
    }
  }
}

class LoopSpecification(override val startSpec: Option[StringNode],
                        override val clauses: List[ContractClause],
                        override val endSpec: Option[StringNode],
                        var srcLocation: SourceLocation,
                       var precedingWhitespace: String="")
  extends generic.syntaxTree.LoopSpecification with LoopSpecOrLabel
object LoopSpecification {
  def unapply(spec: LoopSpecification): Some[(Option[StringNode],
                                        List[ContractClause],
                                        Option[StringNode],
                                        SourceLocation,
                                        String)] = {
    Some(spec.startSpec,
        spec.clauses,
        spec.endSpec,
        spec.srcLocation,
        spec.precedingWhitespace)
  }
}

trait LoopSpecOrLabel extends generic.syntaxTree.LoopSpecOrLabel with VerCorsTreeNode

trait ContractClause extends generic.syntaxTree.ContractClause with VerCorsTreeNode
                      with generic.syntaxTree.LoopClause {
  private var _startSpec: Option[StringNode] = None
  private var _endSpec: Option[StringNode] = None
  def startSpec: Option[StringNode] = _startSpec
  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isVerCorsNode)) {
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a VerCors node")
    }
  }
  def endSpec: Option[StringNode] = _endSpec
  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if(x1.isEmpty || (x1.isDefined && x1.get.isVerCorsNode)) {
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    } else {
      throw new ClassCastException("not a VerCors node")
    }
  }
}

case class NonTranslatedContractClause(var text: String,
                                       var srcLocation: SourceLocation,
                                       var precedingWhitespace: String="")
      extends ContractClause with generic.syntaxTree.NonTranslatedContractClause
      with generic.syntaxTree.NonTranslatedLoopClause

case class Precondition(override val expression: Expression,
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.Precondition with ContractClause

case class Postcondition(override val expression: Expression,
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String="")
      extends generic.syntaxTree.Postcondition with ContractClause

case class Context(expression: Expression,
                   var srcLocation: SourceLocation,
                   var precedingWhitespace: String="")
      extends ContractClause {
  val contextKeyword: generic.syntaxTree.StringNode = ConstantStringNode("context")
  val semicolon: generic.syntaxTree.StringNode = ConstantStringNode(";")

  override def stringRepresentation: String
    = wrappedStringRepresentation(contextKeyword.toString + expression.toString + semicolon.toString)
}

case class ContextEverywhere(expression: Expression,
                             var srcLocation: SourceLocation,
                             var precedingWhitespace: String="")
  extends ContractClause {
  val contextEverywhereKeyword: generic.syntaxTree.StringNode = ConstantStringNode("context_everywhere")
  val semicolon: generic.syntaxTree.StringNode = ConstantStringNode(";")

  override def stringRepresentation: String
    = wrappedStringRepresentation(contextEverywhereKeyword.toString
                                  + expression.toString + semicolon.toString)
}

case class LoopInvariant(override val expression: Expression,
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.LoopInvariant with ContractClause

case class DecreasesClause(override val expression: Expression,
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.DecreasesClause with ContractClause

case class ExceptionalPostcondition(override val exceptionType: TypeIndication,
                                    override val exceptionName: Some[StringNode],
                                    override val expression: Some[Expression],
                                    var srcLocation: SourceLocation,
                                    var precedingWhitespace: String="")
      extends generic.syntaxTree.ExceptionPostcondition with ContractClause

trait GhostIO extends generic.syntaxTree.GhostIO with ContractClause {
  override val argType: TypeIndication
  override val name: StringNode
}
case class GhostArgument(override val argType: TypeIndication,
                         override val name: StringNode,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="")
  extends generic.syntaxTree.GhostArgument with GhostIO

case class GhostReturn(override val argType: TypeIndication,
                         override val name: StringNode,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="")
  extends generic.syntaxTree.GhostReturn with GhostIO


case class NonTranslatedStatement(var text: String,
                                  var srcLocation: SourceLocation,
                                  var precedingWhitespace: String="")
      extends generic.syntaxTree.NonTranslatedStatement with Statement

case class AssertStatement (override val content: Expression,
                            override val message: Option[Expression],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.AssertStatement with Statement
case class AssumeStatement (override val content: Expression,
                            override val message: Option[Expression],
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.AssumeStatement with Statement
case class FoldStatement(override val expression: Expression,
                        var srcLocation: SourceLocation,
                        var precedingWhitespace: String="")
      extends generic.syntaxTree.FoldStatement with Statement
case class UnfoldStatement(override val expression: Expression,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="")
      extends generic.syntaxTree.UnfoldStatement with Statement
case class GhostStatement(statement: Statement,
                          var srcLocation: SourceLocation,
                          var precedingWhitespace: String = "")
      extends generic.syntaxTree.GhostStatement with Statement {
  private var _ghostKeyword: GhostModifier = GhostModifier(srcLocation)

  def ghostKeyword: GhostModifier = _ghostKeyword

  def ghostKeyword_=(x1: generic.syntaxTree.GhostModifier): Unit = {
    x1 match {
      case g: GhostModifier => _ghostKeyword = g
      case _ => throw new ClassCastException("Not a VerCors node")
    }
  }
}

case class JmlStatementBlock(private var _startSpec: Option[StringNode],
                             override val statements: List[Statement],
                             private var _endSpec: Option[StringNode],
                             var srcLocation: SourceLocation,
                             var precedingWhitespace: String="")
      extends generic.syntaxTree.JmlStatementBlock with Statement {
  def startSpec: Option[StringNode] = _startSpec
  def startSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if (x1.isDefined && !x1.get.isInstanceOf[StringNode]) {
      throw new ClassCastException("not a VerCors node")
    }else {
      _startSpec = x1.asInstanceOf[Option[StringNode]]
    }
//    x1 match {
//      case start: Option[StringNode] => startSpec = start
//      case _ => throw new ClassCastException("not a VerCors node")
//    }
  }
  def endSpec: Option[StringNode] = _endSpec
  def endSpec_=(x1: Option[generic.syntaxTree.StringNode]): Unit = {
    if (x1.isDefined && !x1.get.isInstanceOf[StringNode]) {
      throw new ClassCastException("not a VerCors node")
    }else {
      _endSpec = x1.asInstanceOf[Option[StringNode]]
    }
//    x1 match {
////      case end: Option[StringNode] => endSpec = end
////      case _ => throw new ClassCastException("not a VerCors node")
//      case None => endSpec = None
//      case Some(value) => value match {
//        case str: StringNode => endSpec = Some(str)
//        case _ => throw new ClassCastException("not a VerCors node")
//      }
//    }
  }
}

case class ResultExpression(var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.ResultExpression with Expression

case class VerCorsSpecialVariable(var text: String,
                                  var srcLocation: SourceLocation,
                                  var precedingWhitespace: String="")
      extends generic.syntaxTree.StringNode with Expression

case class OldExpression(override val expression: Expression,
                         override val label: Option[StringNode],
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="")
      extends generic.syntaxTree.OldExpression with Expression {
  val openBracket: generic.syntaxTree.StringNode = ConstantStringNode("[")
  val closeBracket: generic.syntaxTree.StringNode = ConstantStringNode("]")

  override def stringRepresentation: String = {
    var res = oldKeyword.toString
    if (label.nonEmpty) {
      res += openBracket.toString
      res += label.get.toString
      res += closeBracket.toString
    }
    res += openParenthesis.toString
    res += expression.toString
    res += closeParenthesis.toString
    res
  }


}

case class TypeOfExpression(override val expression: Expression,
                                 var srcLocation: SourceLocation,
                                 var precedingWhitespace: String="")
      extends generic.syntaxTree.TypeOfExpression with Expression

case class TypeExpression(override val typeIndication: TypeIndication,
                                 var srcLocation: SourceLocation,
                                 var precedingWhitespace: String="")
      extends generic.syntaxTree.TypeExpression with Expression

trait QuantifiedExpression extends generic.syntaxTree.QuantifiedExpression with Expression
case class ForallExpression(override val declaration: List[QuantifiedVarDeclaration],
                            override val guard: Option[Expression],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.ForallExpression with QuantifiedExpression
case class ForallStarExpression(override val declaration: List[QuantifiedVarDeclaration],
                            override val guard: Option[Expression],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.ForallExpression with QuantifiedExpression {
  override val quantifier: syntaxTree.StringNode = ConstantStringNode("\\forall*")
}
case class ExistsExpression(override val declaration: List[QuantifiedVarDeclaration],
                            override val guard: Option[Expression],
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.ExistsExpression with QuantifiedExpression

case class LetExpression(override val declType: TypeIndication,
                         override val declaration: VariableDeclarator,
                            override val expression: Expression,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String="")
      extends generic.syntaxTree.LetExpression with Expression {
}

class QuantifiedVarDeclaration(override val declType: TypeIndication,
                                    override val declarators: List[StringNode],
                                    var srcLocation: SourceLocation,
                                    var precedingWhitespace: String="")
      extends generic.syntaxTree.QuantifiedVarDeclaration with VerCorsTreeNode
object QuantifiedVarDeclaration {
  def unapply(varDecl: QuantifiedVarDeclaration): Some[(TypeIndication,
                                                        List[StringNode],
                                                        SourceLocation,
                                                        String)] = {
    Some(varDecl.declType,
          varDecl.declarators,
          varDecl.srcLocation,
          varDecl.precedingWhitespace)
  }
}

case class PermissionPredicate(override val target: Expression,
                               override val amount: Some[Expression],
                               var srcLocation: SourceLocation,
                               var precedingWhitespace: String="")
  extends generic.syntaxTree.PermissionExpression with Expression {
  val permKeyword: generic.syntaxTree.StringNode = ConstantStringNode("Perm")
  val openParenthesis: generic.syntaxTree.StringNode = ConstantStringNode("(")
  val comma: generic.syntaxTree.StringNode = ConstantStringNode(",")
  val closeParenthesis: generic.syntaxTree.StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    permKeyword.toString + openParenthesis.toString
      + target.toString + comma.toString
      + amount.value.toString + closeParenthesis.toString
  }
}

case class ValuedPermissionPredicate(override val target: Expression,
                                     override val value: Expression,
                                     override val amount: Some[Expression],
                                     var srcLocation: SourceLocation,
                                     var precedingWhitespace: String="")
  extends generic.syntaxTree.ValuedPermissionExpression with Expression {
  val pointsToKeyword: generic.syntaxTree.StringNode = ConstantStringNode("PointsTo")
  val openParenthesis: generic.syntaxTree.StringNode = ConstantStringNode("(")
  val firstComma: generic.syntaxTree.StringNode = ConstantStringNode(",")
  val secondComma: generic.syntaxTree.StringNode = ConstantStringNode(",")
  val closeParenthesis: generic.syntaxTree.StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    pointsToKeyword.toString + openParenthesis.toString
      + target.toString + firstComma.toString
      + value.toString + secondComma.toString
      + amount.value.toString + closeParenthesis.toString
  }
}

case class CurrentPermissionExpression(override val target: Expression,
                                       var srcLocation: SourceLocation,
                                       var precedingWhitespace: String="")
  extends generic.syntaxTree.CurrentPermissionExpression with Expression {
  val permKeyword: generic.syntaxTree.StringNode = ConstantStringNode("perm")
  val openParenthesis: generic.syntaxTree.StringNode = ConstantStringNode("(")
  val closeParenthesis: generic.syntaxTree.StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    permKeyword.toString + openParenthesis.toString
      + target.toString + closeParenthesis.toString
  }
}

case class SomePermissionPredicate(override val target: Expression,
                                   var srcLocation: SourceLocation,
                                   var precedingWhitespace: String="")
  extends generic.syntaxTree.SomePermissionPredicate with Expression {
  val valueKeyword: generic.syntaxTree.StringNode = ConstantStringNode("Value")
  val openParenthesis: generic.syntaxTree.StringNode = ConstantStringNode("(")
  val closeParenthesis: generic.syntaxTree.StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    valueKeyword.toString + openParenthesis.toString
      + target.toString + closeParenthesis.toString
  }
}

case class ArrayExpression(name: Expression,
                           size: Expression,
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String="")
  extends Expression {
  val arrayKeyword: generic.syntaxTree.StringNode = ConstantStringNode("\\array")
  val openParenthesis: generic.syntaxTree.StringNode = ConstantStringNode("(")
  val comma: generic.syntaxTree.StringNode = ConstantStringNode(",")
  val closeParenthesis: generic.syntaxTree.StringNode = ConstantStringNode(")")

  override def stringRepresentation: String = {
    arrayKeyword.toString + openParenthesis.toString + name.toString
      + comma.toString + size.toString + closeParenthesis.toString
  }
}

case class ScaledPredicateExpression(override val amount: Expression,
                                     override val predicate: Expression,
                                     var srcLocation: SourceLocation,
                                     var precedingWhitespace: String="")
  extends generic.syntaxTree.ScaledPredicateExpression with Expression

case class InlinePattern(openPattern: StringNode,
                         expression: Expression,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="")
  extends Expression {
  val closePattern: syntaxTree.StringNode = ConstantStringNode(":}")

  override def stringRepresentation: String = {
    openPattern.toString + expression.toString + closePattern.toString
  }
}

case class UnfoldingExpression(override val predicate: Expression,
                               override val expression: Expression,
                               var srcLocation: SourceLocation,
                               var precedingWhitespace: String="")
  extends generic.syntaxTree.UnfoldingExpression with Expression

case class CardinalityExpression(override val expression: Expression,
                         var srcLocation: SourceLocation,
                         var precedingWhitespace: String="")
  extends generic.syntaxTree.CardinalityExpression with Expression
  
case class OptionSomeExpression(override val expression: Expression,
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.OptionSomeExpression with Expression

case class VerCorsSetComprehension(elementType: TypeIndication,
                                   expression: Expression,
                                   declarations: List[QuantifiedVarDeclaration],
                                   filter: Expression,
                                   var srcLocation: SourceLocation,
                                   var precedingWhitespace: String="")
      extends Expression {
  val setKeyword: StringNode = StringNode("set", srcLocation)
  val openAngle: StringNode = StringNode("<", srcLocation)
  val closeAngle: StringNode = StringNode(">", srcLocation)
  val openBrace: StringNode = StringNode("{", srcLocation)
  val bar: StringNode = StringNode("|", srcLocation)
  val semicolon: StringNode = StringNode(";", srcLocation)
  val closeBrace: StringNode = StringNode("}", srcLocation)

  override def stringRepresentation: String = {
    setKeyword.toString + openAngle.toString + elementType.toString + closeAngle.toString
      + openBrace.toString + expression.toString + bar.toString
      + declarations.mkString(",") + semicolon.toString
      + filter.toString + closeBrace.toString
  }

}

case class TypedSeqLiteral(override val innerType: TypeIndication,
                           override val values: List[Expression],
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String="")
      extends generic.syntaxTree.TypedSeqLiteral with Expression
case class TypedSetLiteral(override val innerType: TypeIndication,
                           override val values: List[Expression],
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String="")
      extends generic.syntaxTree.TypedSetLiteral with Expression
case class TypedBagLiteral(override val innerType: TypeIndication,
                           override val values: List[Expression],
                           var srcLocation: SourceLocation,
                           var precedingWhitespace: String="")
      extends generic.syntaxTree.TypedBagLiteral with Expression

class SeparatingConjunctOperator(pSrcLocation: SourceLocation,
                                 pPrecedingWhitespace: String="") 
      extends StringNode("**", pSrcLocation, pPrecedingWhitespace)
object SeparatingConjunctOperator{
  def unapply(op: SeparatingConjunctOperator): Some[(SourceLocation, String)] = {
    Some(op.srcLocation, op.precedingWhitespace)
  }
}
class ImpliesOperator(pSrcLocation: SourceLocation,
                      pPrecedingWhitespace: String="")
  extends StringNode("==>", pSrcLocation, pPrecedingWhitespace)
object ImpliesOperator{
  def unapply(op: ImpliesOperator): Some[(SourceLocation, String)] = {
    Some(op.srcLocation, op.precedingWhitespace)
  }
}
class PermissionFractionOperator(pSrcLocation: SourceLocation,
                                 pPrecedingWhitespace: String="")
  extends StringNode("\\", pSrcLocation, pPrecedingWhitespace)
object PermissionFractionOperator{
  def unapply(op: PermissionFractionOperator): Some[(SourceLocation, String)] = {
    Some(op.srcLocation, op.precedingWhitespace)
  }
}

case class SpecDeclarationBlock(override val startSpec: Option[StringNode],
                                override val declarations: List[SpecDeclaration],
                                override val endSpec: Option[StringNode],
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.SpecDeclarationBlock with MemberDeclaration()
trait SpecDeclaration extends generic.syntaxTree.SpecDeclaration with MemberDeclaration

case class NonTranslatedSpecDeclaration(var text: String,
                                        var srcLocation: SourceLocation,
                                        var precedingWhitespace: String = "")
      extends generic.syntaxTree.NonTranslatedSpecDeclaration with SpecDeclaration with MemberDeclaration()

case class GhostDeclaration(override val ghost: GhostModifier,
                            override val decl: MemberDeclaration,
                            var srcLocation: SourceLocation,
                            var precedingWhitespace: String = "")
      extends generic.syntaxTree.GhostDeclaration with SpecDeclaration with MemberDeclaration()

case class SpecMethodDeclaration(pName: StringNode,
                                 pReturnType: Option[TypeIndication],
                                 pParameters: List[Parameter],
                                 pContractWithin: Option[Contract],
                                 pBody: BlockOrSemicolon,
                                 pSrcLocation: SourceLocation,
                                 pPrecedingWhitespace: String="")
      extends MethodDeclaration(pName, pReturnType, pParameters, None,
                                None, None, pContractWithin, pBody,
                                pSrcLocation, pPrecedingWhitespace)
        with SpecDeclaration with generic.syntaxTree.SpecMethodDeclaration {

  override def stringRepresentation: String = {
    var res = modifiers.filter(x => x.isInstanceOf[GhostModifier] || x.isInstanceOf[SpecMethodModifier]).mkString
    res += contractBefore.getOrElse("").toString
    res += modifiers.filter(x => !x.isInstanceOf[GhostModifier] && !x.isInstanceOf[SpecMethodModifier]).mkString
    res += returnType.getOrElse("").toString
    res += name.toString
    res += openParenthesis.toString
    res += parameters.mkString(",")
    res += closeParenthesis.toString
    res += dimensions.getOrElse("").toString
    res += throws.getOrElse("").toString
    res += default.getOrElse("").toString
    res += contractWithin.getOrElse("").toString
    res += body.toString
    res
  }


}

case class PredicateDeclaration(override val name: StringNode,
                                override val params: List[Parameter],
                                override val expr: Option[Expression],
                                var srcLocation: SourceLocation,
                                var precedingWhitespace: String="")
      extends generic.syntaxTree.PredicateDeclaration with SpecDeclaration with MemberDeclaration() {
  override val tType: TypeIndication = StringTypeIndication("resource", Nil, "resource", srcLocation)
}