// This example was written by Viktorio S. el Hakim - 1/2021
// It establishes that the array is sorted, but not that it is a permutation of the input array

import java.util.Arrays;

public class MergeSort {    
    /*@requires arr!=null;
       
      requires arr != null && arr.length > 0; //non-null
      ensures (\forall int k; 0<= k && k < arr.length; (\forall int j; k < j && j < arr.length; arr[k] >= arr[j]));
      */
    public static void sort(int [] arr) {
        if (arr.length > 0) {
            sortRec(arr,0,arr.length);
        }
    }
    
    
    /*@requires arr!=null;
       
      requires 0 <= l && l < r && r <= arr.length; // bounds
      requires arr != null; // non-null
      ensures (\forall int k; l <= k && k < r; (\forall int j; k < j && j < r; arr[k] >= arr[j]));
      */
    private static void sortRec(int [] arr, int l, int r) {
        if (l + 1 < r) {
            int mid = l + (r-l)/2;
            sortRec(arr,l,mid);
            sortRec(arr,mid,r);
            merge(arr,l,mid,r);
        }
    }
    
    
    /*@requires arr!=null;
      
      requires 0 <= l && l < m && m < r && r <= arr.length && arr.length > 1; //bounds
      requires (\forall int k, int j; l <= k && k < j && j < m; arr[k] >= arr[j]);
      requires (\forall int k, int j; m <= k && k < j && j < r; arr[k] >= arr[j]);
      ensures  (\forall int k, int j; l <= k && k < j && j < r; arr[k] >= arr[j]) ;
     */
    private static void merge( int [] arr, final int l, final int m, final int r) {
        final /*  non_null */ int [] lCpy = Arrays.copyOfRange(arr, l, m),
                     rCpy = Arrays.copyOfRange(arr,m, r);
            
        //@ assert (\forall int k; 0 <= k && k < lCpy.length; lCpy[k] == arr[k+l]);
        //@ assert (\forall int k; 0 <= k && k < rCpy.length; rCpy[k] == arr[k+m]);
        
        int l1 = 0, r1 = 0;
        /*@
         loop_invariant Perm( i, write) ** Perm( l1, write) ** Perm( r1, write) **(\forall* int i0;0 <= i0 && i0 < arr.length; Perm( arr[ i0], write));
         loop_invariant 0 <= l1 && l1 <= lCpy.length && 0 <= r1 && r1 <= rCpy.length;
         loop_invariant l <= i && i <= r && i == l+l1+r1;
         loop_invariant (\forall int k, int j; l <= k && k < i && l1 <= j && j < lCpy.length; arr[k] >= lCpy[j]);
         loop_invariant (\forall int k, int j; l <= k && k < i && r1 <= j && j < rCpy.length; arr[k] >= rCpy[j]);
         loop_invariant (\forall int k, int j; l <= k && k < j && j < i; arr[k] >= arr[j]);
         decreases r - i;
         */
        for (int i=l; i < r; i++) {
            if (l1 < lCpy.length && (r1 >= rCpy.length || lCpy[l1] >= rCpy[r1])) {
                arr[i] = lCpy[l1++];
            } else {
                arr[i] = rCpy[r1++];
            }
        }
    }
}
