// Issue: MIN_VALUE and MAX_VALUE cannot be equal to the actual MIN and MAX values.
public class SumMax {
  int MIN_VALUE;
  int MAX_VALUE;
  
  SumMax() {
    MIN_VALUE = -2147483647; // off-by-one, see issue #640
    MAX_VALUE = 2147483646; // off-by-one, see issue #640
  }

  //@ requires a != null;
  //@ requires a.length > 0;
  //@ requires (\forall* int i; i >= 0 && i < a.length; Perm(a[i], read));
  //@ requires Perm(MIN_VALUE, read);
  //@ requires Perm(MAX_VALUE, read);
  //@ ensures Perm(MAX_VALUE, read);
  //@ ensures Perm(MIN_VALUE, read);
  //@ ensures (\forall* int i; i >= 0 && i < a.length; Perm(a[i], read));
  void m(int[] a) {
    int sum = 0;
    int max = a[0];

    //@ loop_invariant (\forall* int i; i >= 0 && i < a.length; Perm(a[i], read));
    //@ loop_invariant Perm(MIN_VALUE, read);
    //@ loop_invariant Perm(MAX_VALUE, read);
    //@ loop_invariant 0 <= i && i <= a.length;
    //@ loop_invariant sum <= i * max;  // Assertion to be proved
    for (int i=0; i<a.length; i++) {
      //@ assume MIN_VALUE <= sum + a[i] && sum + a[i] <= MAX_VALUE; // Just assume we never overflow
      sum += a[i];
      if (max < a[i]) max = a[i];
    }
  }

}

