import intermediateRepresentation.IRProgram
import languages.Krakatoa.KrakatoaProgram
import languages.OpenJML.OpenJMLProgram
import org.scalatest.flatspec.AnyFlatSpec

import java.io.{File, PrintWriter}
import scala.io.Source

class KrakatoaParserTest extends AnyFlatSpec {

  def parseKrakatoaFile(filename: String): KrakatoaProgram = {
    val testFileContents = Source.fromURL(getClass.getResource(filename)).mkString //NOTE: This leaves the file open as long as the JVM is running!
    KrakatoaProgram.fromText(testFileContents)
  }
  
  "The AnnotationTranslator" should "parse Krakatoa annotations in ArrayMax.java" in {
    val res = parseKrakatoaFile("/Krakatoa/ArrayMax.java")
    testParsing("ArrayMax.java", res)
  }

  it should "parse Krakatoa annotations in BinarySearch.java" in {
    val res = parseKrakatoaFile("/Krakatoa/BinarySearch.java")
    testParsing("BinarySearch.java", res)
  }

  it should "parse the Krakatoa annotations in Deterrence.java" in {
    val res = parseKrakatoaFile("/Krakatoa/Deterrence.java")
    testParsing("Deterrence.java", res)
  }

  it should "parse the Krakatoa annotations in Duplets.java" in {
    val res = parseKrakatoaFile("/Krakatoa/Duplets.java")
    testParsing("Duplets.java", res)
  }

  it should "parse the Krakatoa annotations in Fibonacci.java" in {
    val res = parseKrakatoaFile("/Krakatoa/Fibonacci.java")
    testParsing("Fibonacci.java", res)
  }

  it should "parse the Krakatoa annotations in FlagStatic.java" in {
    val res = parseKrakatoaFile("/Krakatoa/FlagStatic.java")
    testParsing("FlagStatic.java", res)
  }

  it should "parse the Krakatoa annotations in Gcd.java" in {
    val res = parseKrakatoaFile("/Krakatoa/Gcd.java")
    testParsing("Gcd.java", res)
  }

  it should "parse the Krakatoa annotations in MacCarthy.java" in {
    val res = parseKrakatoaFile("/Krakatoa/MacCarthy.java")
    testParsing("MacCarthy.java", res)
  }

  it should "parse the Krakatoa annotations in Muller.java" in {
    val res = parseKrakatoaFile("/Krakatoa/Muller.java")
    testParsing("Muller.java", res)
  }

  it should "parse the Krakatoa annotations in Purse.java" in {
    val res = parseKrakatoaFile("/Krakatoa/Purse.java")
    testParsing("Purse.java", res)
  }

  it should "parse the Krakatoa annotations in Sort.java" in {
    val res = parseKrakatoaFile("/Krakatoa/Sort.java")
    testParsing("Sort.java", res)
  }

  it should "parse the Krakatoa annotations in Tree.java" in {
    val res = parseKrakatoaFile("/Krakatoa/Tree.java")
    testParsing("Tree.java", res)
  }

  it should "parse a Krakatoa file in String format" in {
    val input="/**************************************************************************/\n/*                                                                        */\n/*  The Why platform for program certification                            */\n/*                                                                        */\n/*  Copyright (C) 2002-2011                                               */\n/*                                                                        */\n/*    Jean-Christophe FILLIATRE, CNRS & Univ. Paris-sud 11                */\n/*    Claude MARCHE, INRIA & Univ. Paris-sud 11                           */\n/*    Yannick MOY, Univ. Paris-sud 11                                     */\n/*    Romain BARDOU, Univ. Paris-sud 11                                   */\n/*                                                                        */\n/*  Secondary contributors:                                               */\n/*                                                                        */\n/*    Thierry HUBERT, Univ. Paris-sud 11  (former Caduceus front-end)     */\n/*    Nicolas ROUSSET, Univ. Paris-sud 11 (on Jessie & Krakatoa)          */\n/*    Ali AYAD, CNRS & CEA Saclay         (floating-point support)        */\n/*    Sylvie BOLDO, INRIA                 (floating-point support)        */\n/*    Jean-Francois COUCHOT, INRIA        (sort encodings, hyps pruning)  */\n/*    Mehdi DOGGUY, Univ. Paris-sud 11    (Why GUI)                       */\n/*                                                                        */\n/*  This software is free software; you can redistribute it and/or        */\n/*  modify it under the terms of the GNU Lesser General Public            */\n/*  License version 2.1, with the special exception on linking            */\n/*  described in file LICENSE.                                            */\n/*                                                                        */\n/*  This software is distributed in the hope that it will be useful,      */\n/*  but WITHOUT ANY WARRANTY; without even the implied warranty of        */\n/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  */\n/*                                                                        */\n/**************************************************************************/\n\n//@+ CheckArithOverflow = no\n\n\nclass NoCreditException extends Exception {\n\n    public NoCreditException() { }\n\n}\n\npublic class Purse {\n\n    private int balance;\n    //@ invariant balance_non_negative: balance >= 0;\n\n    /*@ requires true;\n      @ assigns balance;\n      @ ensures balance == 0;\n      @*/\n    public Purse() {\n        balance = 0;\n    }\n\n    /*@ requires s >= 0;\n      @ assigns balance;\n      @ ensures balance == \\old(balance) + s;\n      @*/\n    public void credit(int s) {\n        balance += s;\n    }\n\n    /*@ requires s >= 0;\n      @ assigns balance;\n      @ ensures s <= \\old(balance) && balance == \\old(balance) - s;\n      @ behavior amount_too_large:\n      @   assigns \\nothing;\n      @   signals (NoCreditException) s > \\old(balance) ;\n      @*/\n    public void withdraw(int s) throws NoCreditException {\n        if (balance >= s)\n            balance = balance - s;\n        else\n            throw new NoCreditException();\n    }\n\n    //@ ensures \\result == balance;\n    public int getBalance() {\n        return balance;\n    }\n\n}\n\n\n\n/*\nLocal Variables:\ncompile-command: \"make Purse.why3ml\"\nEnd:\n*/"
    KrakatoaProgram.fromText(input)
  }

  private val inputDir = "./src/test/resources/Krakatoa/"
//  private val outDir = inputDir + "out/"
//  private val testFiles = List("ArrayMax.java",
//    "BinarySearch.java",
//    "Deterrence.java",
//    "Duplets.java",
//    "Fibonacci.java",
//    "FlagStatic.java",
//    "Gcd.java",
//    "MacCarthy.java",
//    "Muller.java",
//    "Purse.java",
//    "Sort.java",
//    "Tree.java")
//
//  for(filename <- testFiles) {
//    val krakatoa = parseKrakatoaFile(inputDir+filename)
//    s"$filename" should s"give expected content" in {
//      testParsing(filename, krakatoa)
//    }
////    s"$filename" should s"be unchanged when translating to IR and back" in {
////      testIRTranslation(krakatoa)
////    }
//  }
//  def translateKrakatoaToOpenJML(krakatoa: KrakatoaProgram): String = {
//    val ir = krakatoa.toIRProgram
//    val openjml = OpenJMLProgram.fromIRProgram(ir)
//    openjml.toText
//  }

  def testParsing(filename: String, program: KrakatoaProgram): Unit = {
    val expectedIterator = Source.fromFile(inputDir+filename)
    try {
      val expectedContent = expectedIterator.getLines().mkString("\n")
      assert(program.toText == expectedContent)
    } finally {
      expectedIterator.close()
      System.out.flush()
      System.err.flush()
    }
  }

//  def testIRTranslation(program: KrakatoaProgram): Unit = {
//    val irProgram: IRProgram = program.toIRProgram
//    val reconstructed: KrakatoaProgram = KrakatoaProgram.fromIRProgram(irProgram)
//    assert(program.toText == reconstructed.toText)
//  }

}
