import intermediateRepresentation.IRProgram
import languages.Krakatoa.KrakatoaProgram
import languages.VerCors.VerCorsProgram
import org.scalatest.flatspec.AnyFlatSpec

import java.io.{File, PrintWriter}
import java.nio.file.Paths
import scala.io.Source

class KrakatoaToVerCorsTest extends AnyFlatSpec {
  private val sep = Paths.get(".").getFileSystem.getSeparator
  private val inputPath = List(".","src","test","resources","Krakatoa")
  private val outputPath = inputPath.appendedAll(List("out", "VerCors"))
  private val testFiles = List("ArrayMax.java",
    "BinarySearch.java",
    "Deterrence.java",
    "Duplets.java",
    "Fibonacci.java",
    "FlagStatic.java",
    "Gcd.java",
    "MacCarthy.java",
    "Muller.java",
    "Purse.java",
    "Sort.java",
    "Tree.java")

  def inputFilePath(filename: String): String = {
    inputPath.appended(filename).mkString(sep)
  }

  def outputFilePath(filename: String): String = {
    outputPath.appended(filename).mkString(sep)
  }

  def printToFile(filename: String, contents: String) = {
    val p = new PrintWriter(new File(outputFilePath(filename)))
    try {
      p.write(contents)
    } finally {
      p.close()
    }
  }

  "The AnnotationTranslator" should "translate ArrayMax's annotations to VerCors's specification language" in {
    val filename = "ArrayMax.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate BinarySearch's annotations to VerCors's specification language" in {
    val filename = "BinarySearch.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate Deterrence's annotations to VerCors's specification language" in {
    val filename = "Deterrence.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate Duplets's annotations to VerCors's specification language" in {
    val filename = "Duplets.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate Fibonacci's annotations to VerCors's specification language" in {
    val filename = "Fibonacci.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate FlagStatic's annotations to VerCors's specification language" in {
    val filename = "FlagStatic.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate Gcd's annotations to VerCors's specification language" in {
    val filename = "Gcd.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate MacCarthy's annotations to VerCors's specification language" in {
    val filename = "MacCarthy.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate Muller's annotations to VerCors's specification language" in {
    val filename = "Muller.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate Purse's annotations to VerCors's specification language" in {
    val filename = "Purse.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate Sort's annotations to VerCors's specification language" in {
    val filename = "Sort.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  "The AnnotationTranslator" should "translate Tree's annotations to VerCors's specification language" in {
    val filename = "Tree.java"
    val krakatoaProgram = KrakatoaProgram.fromFile(inputFilePath(filename))
    val vercors = testIRTranslation(krakatoaProgram)
    printToFile(filename, vercors.toText)
  }

  def testIRTranslation(krakatoa: KrakatoaProgram): VerCorsProgram = {
    val irProgram: IRProgram = krakatoa.toIRProgram
    VerCorsProgram.fromIRProgram(irProgram)
  }
}
