import intermediateRepresentation.IRProgram
import languages.OpenJML.OpenJMLProgram
import org.scalatest.flatspec.AnyFlatSpec

import java.nio.file.Paths
import scala.io.Source


class OpenJMLParserTest extends AnyFlatSpec {
  private val sep = Paths.get(".").getFileSystem.getSeparator
  private val lineSep = util.Properties.lineSeparator
  private val inputPath = List(".","src","test","resources","OpenJML")
  private val outputPath = inputPath.appended("output")
  private val parsedPath = outputPath.appended("parsed")
  private val reconstructedPath = outputPath.appended("OpenJMLProgram")
  private val testFiles = List(
    "BinarySearch.java",
    "BubbleSort.java",
    "BubbleSortAlso.java",
    "BTest.java",
    "BTest2.java",
    "BTest3.java",
    "BTest4.java",
    "BTestNormalExceptional.java",
    "InvertInjection.java",
    "MaxByElimination.java",
    "MergeSort.java",
    "SelectionSort.java",
    "SumMax.java",
    "VerifyThis2019_2A.java")

  def inputFilePath(filename: String): String = {
    inputPath.appended(filename).mkString(sep)
  }

  def reconstructedFilePath(filename: String): String = {
    reconstructedPath.appended(filename).mkString(sep)
  }
  def parsedFilePath(filename: String): String = {
    parsedPath.appended(filename).mkString(sep)
  }

  for(filename <- testFiles) {
    val openjml = OpenJMLProgram.fromFile(inputFilePath(filename))
    s"$filename" should s"give expected content" in {
      testParsing(filename, openjml)
    }
    s"$filename" should s"be unchanged when translating to IR and back" in {
      testIRTranslation(filename, openjml)
    }
  }

  def testParsing(filename: String, openJMLProgram: OpenJMLProgram): Unit = {
    val expectedIterator = Source.fromFile(parsedFilePath(filename))
    try {
      val expectedContent = expectedIterator.getLines().mkString(lineSep)
      assert(openJMLProgram.toText.stripTrailing() == expectedContent.stripTrailing())
    } finally {
      expectedIterator.close()
    }
  }

  def testIRTranslation(filename: String, openJMLProgram: OpenJMLProgram): Unit = {
    val irProgram: IRProgram = openJMLProgram.toIRProgram
    val reconstructed: OpenJMLProgram = OpenJMLProgram.fromIRProgram(irProgram)
    val expectedIterator = Source.fromFile(reconstructedFilePath(filename))
    try {
      val expectedContent = expectedIterator.getLines().mkString(lineSep)
      assert(reconstructed.toText.stripTrailing() == expectedContent.stripTrailing())
    } finally {
      expectedIterator.close()
    }
  }
}
