import intermediateRepresentation.IRProgram
import languages.OpenJML.OpenJMLProgram
import languages.VerCors.VerCorsProgram
import org.scalatest.flatspec.AnyFlatSpec

import java.nio.file.Paths
import scala.io.Source

class OpenJMLtoVerCorsTest extends AnyFlatSpec {
  private val sep = Paths.get(".").getFileSystem.getSeparator
  private val lineSep = util.Properties.lineSeparator
  private val inputPath = List(".","src","test","resources","OpenJML")
  private val outputPath = inputPath.appendedAll(List("output","VerCorsProgram"))
  private val testFiles = List(
    "BinarySearch.java",
    "BubbleSort.java",
    "BubbleSortAlso.java",
    "BTest.java",
    "BTest2.java",
    "BTest3.java",
    "BTest4.java",
    "BTestNormalExceptional.java",
    "InvertInjection.java",
    "MaxByElimination.java",
    "MergeSort.java",
    "SelectionSort.java",
    "SumMax.java",
    "VerifyThis2019_2A.java")

  def inputFilePath(filename: String): String = {
    inputPath.appended(filename).mkString(sep)
  }

  def outputFilePath(filename: String): String = {
    outputPath.appended(filename).mkString(sep)
  }

  for(filename <- testFiles) {
    val openjml = OpenJMLProgram.fromFile(inputFilePath(filename))
    s"$filename" should s"give expected content" in {
      testIRTranslation(filename, openjml)
    }
  }

  def testIRTranslation(filename: String, openJMLProgram: OpenJMLProgram): Unit = {
    val irProgram: IRProgram = openJMLProgram.toIRProgram
    val vercorsProgram: VerCorsProgram = VerCorsProgram.fromIRProgram(irProgram)
    assert(vercorsProgram != null)
    val expectedIterator = Source.fromFile(outputFilePath(filename))
    try {
      val expectedContent = expectedIterator.getLines().mkString(lineSep).stripTrailing()
      val actual = vercorsProgram.toText.stripTrailing()
      assert(actual == expectedContent)
    } finally {
      expectedIterator.close()
    }
  }
}
