import intermediateRepresentation.IRProgram
import languages.OpenJML.OpenJMLProgram
import languages.VerCors.VerCorsProgram
import org.scalatest.flatspec.AnyFlatSpec

import java.nio.file.{Path, Paths}
import scala.io.Source

class VerCorsParserTest extends AnyFlatSpec {

  private val sep = Paths.get(".").getFileSystem.getSeparator
  private val lineSep = util.Properties.lineSeparator
  private val inputPath = List(".","src","test","resources","VerCors")
  private val outputPath = inputPath.appended("out")
  private val reconstructedPath = outputPath.appended("VerCorsProgram")
  private val parsedPath = outputPath.appended("parsed")
  private val testFiles = List(
                                "Challenge2A2019.java",
                                "Counter.java",
                                "CounterVerifast.java",
                                "Incr.java",
//                                "JavaArrayExamples.java",
                                "MaxByElimination.java",
                                "SumMax.java"
                                  )

  def inputFilePath(filename: String): String = {
    inputPath.appended(filename).mkString(sep)
  }

  def parsedFilePath(filename: String): String = {
    parsedPath.appended(filename).mkString(sep)
  }
  def reconstructedFilePath(filename: String): String = {
    reconstructedPath.appended(filename).mkString(sep)
  }

  for(filename <- testFiles) {
    val vercors = VerCorsProgram.fromFile(inputFilePath(filename))
    s"$filename" should s"give expected content" in {
      testParsing(filename, vercors)
    }
    s"$filename" should s"be unchanged when translating to IR and back" in {
      testIRTranslation(filename, vercors)
    }
  }

  def testParsing(filename: String, verCorsProgram: VerCorsProgram): Unit = {
    val expectedIterator = Source.fromFile(parsedFilePath(filename))
    try {
      val expectedContent = expectedIterator.getLines().mkString(lineSep)
      assert(verCorsProgram.toText.stripTrailing() == expectedContent.stripTrailing())
    } finally {
      expectedIterator.close()
    }
  }

  def testIRTranslation(filename: String, verCorsProgram: VerCorsProgram): Unit = {
    val irProgram: IRProgram = verCorsProgram.toIRProgram
    val reconstructed: VerCorsProgram = VerCorsProgram.fromIRProgram(irProgram)
    val expectedIterator = Source.fromFile(reconstructedFilePath(filename))
    try {
      val expectedContent = expectedIterator.getLines().mkString(lineSep)
      assert(reconstructed.toText.stripTrailing() == expectedContent.stripTrailing())
    } finally {
      expectedIterator.close()
    }
  }

}
