import intermediateRepresentation.IRProgram
import languages.OpenJML.OpenJMLProgram
import languages.VerCors.VerCorsProgram
import org.scalatest.flatspec.AnyFlatSpec

import java.nio.file.Paths
import scala.io.Source

class VerCorsToOpenJML extends AnyFlatSpec {
  private val sep = Paths.get(".").getFileSystem.getSeparator
  private val lineSep = util.Properties.lineSeparator
  private val inputPath = List(".","src","test","resources","VerCors")
  private val outputPath = inputPath.appendedAll(List("out", "OpenJMLProgram"))
  private val testFiles = List(
    "Challenge2A2019.java",
    "Counter.java",
    "CounterVerifast.java",
    "Incr.java",
//    "JavaArrayExamples.java",
    "MaxByElimination.java",
    "SumMax.java"
  )

  def inputFilePath(filename: String): String = {
    inputPath.appended(filename).mkString(sep)
  }

  def outputFilePath(filename: String): String = {
    outputPath.appended(filename).mkString(sep)
  }

  for(filename <- testFiles) {
    val vc = VerCorsProgram.fromFile(inputFilePath(filename))
    s"$filename" should s"give expected content" in {
      testIRTranslation(filename, vc)
    }
  }

  def testIRTranslation(filename: String, vcProgram: VerCorsProgram): Unit = {
    val irProgram: IRProgram = vcProgram.toIRProgram
    val openjml: OpenJMLProgram = OpenJMLProgram.fromIRProgram(irProgram)
    assert(openjml != null)
    val expectedIterator = Source.fromFile(outputFilePath(filename))
    try {
      val expectedContent = expectedIterator.getLines().mkString(lineSep).stripTrailing()
      val actual = openjml.toText.stripTrailing()
      assert(actual == expectedContent)
    } finally {
      expectedIterator.close()
    }
  }
}
