import sys
import re

vercors_style = False

def get_specs_and_comments(content):
  specs = []
  comments = []
  in_comment = False
  in_spec = False
  for l in content:
    if not l or l.isspace():
      continue
    # slight improvement to before when dealing with multiple /* */ in a line
    # still falls short if block and line comment delimiters are mixed
    if in_spec:
      specs.append(l)
      if "*/" in l:
        in_spec = False
        last = re.split('[*]/', l)[-1]
        if "/*@" in last:
          in_spec = True
        elif re.search("/[*]\s*@", last) and vercors_style:
          in_spec = True
        elif "/*" in last:
          in_comment = True
          comments.append(l)
    elif in_comment:
      if l.strip().startswith("@") and vercors_style:
        specs.append(l)
      else:
        comments.append(l)
      if "*/" in l:
        in_comment = False
        last = re.split('[*]/', l)[-1]
        tail = re.split('[*]/', l, 1)[-1]
        if "/*@" in tail or (re.search("/[*]\s*@", tail) and vercors_style):
          specs.append(l)
        if "/*@" in last or (re.search("/[*]\s*@", last) and vercors_style):
          in_spec = True
        elif "/*" in last:
          in_comment = True
    else:
      # neither in spec block nor in comment block when starting this line
      if "/*@" in l or (re.search("/[*]\s*@",l) and vercors_style):
        specs.append(l)
        last = re.split('/[*]', l)[-1]
        if "*/" not in last:
          if last.startswith("@") or (last.strip().startswith("@") and vercors_style):
            in_spec = True
          else:
            in_comment = True
            comments.append(l)
      elif "/*" in l:
        comments.append(l)
        last = re.split('/[*]', l)[-1]
        if "*/" not in last:
            in_comment = True
      elif "//@" in l.strip():
        specs.append(l)
      elif "//" in l.strip():
        comments.append(l)

    
    # if in_spec and "*/" in l.strip():
      # in_spec = False
      # specs.append(l)
    # elif in_comment and "*/" in l.strip():
      # in_comment = False
      # comments.append(l)
    # elif in_spec and not l.isspace():
      # specs.append(l)
    # elif in_comment and not l.isspace():
      # comments.append(l)
    # elif "//@" in l.strip():
      # specs.append(l)
    # elif l.strip().startswith("//"):
      # comments.append(l)
    # elif "/*@" in l.strip() and "*/" in l.strip():
      # specs.append(l)
    # elif "/*@" in l.strip():
      # in_spec = True
      # specs.append(l)
    # elif re.search('^//[^@]', l.strip()):
      # comments.append(l)
    # elif re.search('^/[*][^@].*[*]/$', l.strip()):
      # comments.append(l)
    # elif re.search('^/[*].*', l.strip()):
      # in_comment = True
      # comments.append(l)
  return specs, comments

def get_code(content):
  code = []
  in_block = False
  for l in content:
    if in_block and "*/" in l.strip() and not l.strip().endswith('*/'):
      in_block = False
      code.append(l)
    elif in_block and l.strip().endswith("*/"):
      in_block = False
      continue
    elif in_block and "*/" not in l.strip():
      continue
    elif l.strip().startswith("//"):
      continue
    elif l.strip().startswith("/*") and "*/" not in l.strip():
      in_block = True
      continue
    elif l.strip().startswith("/*") and l.strip().endswith("*/"):
      continue
    elif l.strip().startswith("/*") and "*/" in l.strip() and not l.strip().endswith("*/"):
      code.append(l)
    elif "/*" in l.strip() and not l.strip().startswith("/*") and "*/" not in l.strip():
      in_block = True
      code.append(l)
    else:
      code.append(l)
  return code

def get_empties(content):
  empties = []
  in_block = False
  for l in content:
    if in_block and l.isspace():
      empties.append(l)
    if in_block and not "*/" in l.strip():
      continue
    elif in_block and "*/" in l.strip():
      in_block = False
    elif "/*@" in l.strip() and "*/" not in l.strip():
      in_block = True
    elif not l:
      empties.append(l)
    elif l.isspace():
      empties.append(l)
    elif l.strip() == '':
      empties.append(l)
  return empties

def print_base_stats(filename):
    with open(filename) as f:
        content = f.readlines()
    lineCount = len(content)
    empties = get_empties(content) # This includes completely empty lines in spec blocks
    emptyCount = len(empties)
    specs, comments = get_specs_and_comments(content)
    specCount = len(specs)
    emptySpecs = [l for l in specs 
                    if l.isspace() 
                      or not l 
                      or l.strip() in ["//@", "//", "*/", "/*@", "@*/", "@", ''] 
                      or l.strip().endswith("/*@") 
                      or (l.strip().startswith("@*/") 
                          and "/*@" not in l and "//@" not in l)
                      or (l.strip().startswith("*/") 
                          and "/*@" not in l and "//@" not in l)
                      or re.fullmatch("@\s*//.*", l.strip()) 
                      or re.fullmatch("/[*]@\s*//.*", l.strip()) 
                      or re.fullmatch("//@\s*//.*", l.strip()) 
                      or re.fullmatch("//[^@].*", l.strip())]
    emptySpecCount = len(emptySpecs)
    commentCount = len(comments)
    
    code = get_code([l for l in content if l not in empties])
    codeCount = len(code)
    codeAndSpec = [l for l in code 
                      if re.search('/[*]@', l) 
                        or re.search('@[*]/', l) 
                        or ("//@" in l.strip() and not l.strip().startswith("//@"))]
    codeAndSpecCount = len(codeAndSpec)
    #print "".join(empties)
    # print("Total number of lines of code: "+str(codeCount))
    # print("Total number of empty lines: "+str(emptyCount))
    # print("Total number of specification lines: "+str(specCount))
    print("empty specs:")
    for l in emptySpecs:
      print(l.rstrip())
    print("Total number of empty specification lines: "+str(emptySpecCount))
    print("----------------------------")
    print("Non-empty specs:")
    for l in [l for l in specs if l not in emptySpecs]:
      print(l.rstrip())
    print("Total number of non-empty specification lines: "+str(specCount - emptySpecCount))
    # print("Total number of specification lines on code lines: "+str(codeAndSpecCount))
    print("----------------------------")
    print("Non-specs:")
    for l in [l for l in content if l not in specs]:
      print(l.rstrip())
    # print("Total number of comments: "+str(commentCount))
    # print("Total number of lines in file: "+str(lineCount))
    return None

args = sys.argv
if "--vercors" in args:
  vercors_style = True
  args.remove("--vercors")
print_base_stats(args[1])

